from collections import defaultdict
from datetime import timedelta
from time import perf_counter, sleep
from typing import Any, Callable, DefaultDict, Dict, Hashable, Iterator, List, Optional, Tuple, TypeVar, Union

from typing_extensions import Self

T = TypeVar("T")


def iter_timer(it: Iterator[T]) -> Iterator[Tuple[T, float]]:
    try:
        while True:
            start = perf_counter()
            res = next(it)
            delta = perf_counter() - start
            yield res, delta
    except StopIteration:
        pass


class TakeAtleast:
    __slots__ = ("delta", "wait_on_error", "now")
    delta: float
    wait_on_error: bool
    now: Optional[float]

    def __init__(self, delta: Union[float, timedelta], wait_on_error: bool = False) -> None:
        """Uses `time.perf_counter` and `os.sleep` to make sure the execution of the code under context manager
        took at least `delta` time. If the code executed faster it will simply sleep afterwards.
        If `wait_on_error` is False and an exception is thrown, it won't sleep.
        """

        if isinstance(delta, timedelta):
            self.delta = delta.total_seconds()
        else:
            self.delta = float(delta)

        self.wait_on_error = wait_on_error
        self.now = None

    def __enter__(self) -> Self:
        self.now = perf_counter()
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> Optional[bool]:
        if exc_type is None or self.wait_on_error:
            elapsed = perf_counter() - self.now
            if elapsed < self.delta:
                sleep(self.delta - elapsed)


class DeltaTime:
    __slots__ = ("start", "end")
    start: Optional[float]
    end: float

    def __init__(self) -> None:
        self.start = None
        self.end = perf_counter()

    def __iter__(self) -> Self:
        return self

    def __next__(self) -> float:
        self.start, self.end = self.end, perf_counter()
        return self.end - self.start

    def get(self) -> float:
        return perf_counter() - self.end

    def reset(self) -> None:
        self.end = perf_counter()

    def get_reset(self) -> float:
        self.start, self.end = self.end, perf_counter()
        return self.end - self.start


class PrintStatementTime:
    __slots__ = ("tpl", "start", "delta", "interrupted")
    tpl: str
    start: Optional[float]
    delta: Optional[float]
    interrupted: Optional[bool]

    def __init__(self, tpl: Optional[str] = None) -> None:
        """Times the execution of the code under the context manager and print it afterwards.
        `tpl` is a format string with one field `delta` to include time in seconds.
        The default is `"Execution took {delta}s"`.
        """

        if tpl is None:
            self.tpl = "Execution took {delta}s"
        else:
            self.tpl = tpl
        self.start = None
        self.delta = None
        self.interrupted = None

    def __enter__(self) -> Self:
        self.start = perf_counter()
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> Optional[bool]:
        self.delta = perf_counter() - self.start
        msg = self.tpl.format(delta=self.delta)
        if exc_type:
            self.interrupted = True
            print(msg + " (interrupted)")
        else:
            self.interrupted = False
            print(msg)


class MeasureTime:
    __slots__ = ("delta", "start", "interrupted")
    delta: Optional[float]
    start: Optional[float]
    interrupted: Optional[bool]

    def __init__(self) -> None:
        self.delta = None
        self.start = None
        self.interrupted = None

    def __enter__(self) -> Self:
        self.start = perf_counter()
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> Optional[bool]:
        self.delta = perf_counter() - self.start
        if exc_type:
            self.interrupted = True
        else:
            self.interrupted = False

    def get(self) -> float:
        if self.start is None:
            raise RuntimeError("Context not entered")

        if self.delta is None:
            return perf_counter() - self.start

        return self.delta


class TimeIt:
    def __init__(self) -> None:
        self.results: DefaultDict[Hashable, List[float]] = defaultdict(list)
        self.starts: Dict[Hashable, DeltaTime] = {}

    def __call__(self, key: Hashable, func: Callable[..., T], *args: Any, **kwargs: Any) -> T:
        with MeasureTime() as t:
            ret = func(*args, **kwargs)
        self.results[key].append(t.get())
        return ret

    def start(self, key: Hashable) -> None:
        self.starts[key] = DeltaTime()

    def stop(self, key: Hashable) -> None:
        self.results[key].append(self.starts[key].get())

    def min(self, key: Hashable) -> float:
        return min(self.results[key])

    def length(self, key: Hashable) -> int:
        return len(self.results[key])



import csv
from datetime import datetime
from decimal import Decimal
from itertools import chain, repeat
from operator import itemgetter
from typing import Any, Iterator, Optional

from .callbacks import Progress
from .dict import mapmap
from .exceptions import InconsistentState, NoResult
from .file import copen
from .sqlite import quote_identifier
from .string import build_multiple_replace
from .typing import Connection, Cursor

sqltypes_to_str = {
    str: "str",
    datetime: "datetime",
    int: "int",
    Decimal: "decimal",
}

str_to_sqlitetype = {
    "str": "TEXT",
    "datetime": "TEXT",
    "int": "INTEGER",
    "decimal": "REAL",
}


_convert_mysql_to_sqlite = build_multiple_replace({"\\\\": "\\", "\\'": "''"})


def convert_mysql_to_sqlite(s: str) -> str:
    return _convert_mysql_to_sqlite(s)


class TransactionCursor:
    """Cursor context manager which starts a transaction and rolls back in case of error."""

    def __init__(self, conn: Connection) -> None:
        self.cursor = conn.cursor()

    def __enter__(self) -> Cursor:
        self.cursor.execute("BEGIN TRANSACTION")
        return self.cursor

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        if exc_type:
            self.cursor.execute("ROLLBACK TRANSACTION")
        else:
            self.cursor.execute("COMMIT TRANSACTION")
        self.cursor.close()


class CursorContext:
    """Cursor context manager which creates a new cursor and closes it when it leaves the context."""

    def __init__(self, conn: Connection) -> None:
        self.cursor = conn.cursor()

    def __enter__(self) -> Cursor:
        return self.cursor

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.cursor.close()


def upsert(cursor: Cursor, primary: dict, values: dict, table: str) -> bool:
    """Inserts `table` fields specified by `primary` and `values` keys,
    with the corresponding values.
    If all the values specified by `primary` already exist,
    they are updated instead.

    `table`, `primary.keys()` and `values.keys()` are not escaped.
    `primary.values()` and `values.values()` are escaped.
    """

    # use INSERT ... ON DUPLICATE KEY UPDATE instead?

    if not primary:
        raise ValueError("Empty primary mapping would result in an empty WHERE condition which would affect all rows")

    set_str = ",".join(f"{k}=?" for k in values.keys())
    where_str = " AND ".join(f"{k}=?" for k in primary.keys())

    cursor.execute(f"UPDATE {table} SET {set_str} WHERE {where_str}", chain(values.values(), primary.values()))  # nosec

    if cursor.rowcount == 0:
        into_str = ",".join(chain(primary.keys(), values.keys()))
        values_str = ",".join(repeat("?", len(primary) + len(values)))
        cursor.execute(
            f"INSERT INTO {table} ({into_str}) VALUES ({values_str})",  # nosec
            chain(primary.values(), values.values()),
        )
        return True

    return False


def fetchone(cursor: Cursor) -> Any:
    """Fetch results from `cursor` and assure only one result was returned."""

    rows = cursor.fetchall()
    if len(rows) == 0:
        raise NoResult("No result found")
    elif len(rows) == 1:
        return rows[0]
    else:
        raise InconsistentState("More than one result found")


def iterfetch(cursor: Cursor, batchsize: int = 1000) -> Iterator[Any]:
    """Iterate all results from `cursor`."""

    while True:
        results = cursor.fetchmany(batchsize)
        if not results:
            break
        yield from results


def export_sql_to_csv(
    connection: Connection, path: str, query: str, queryargs: tuple = (), progress: Optional[Progress] = None
) -> None:
    """Exports the result of `query` from a SQL database to a csv file `path`.
    `queryargs` will be passed to the query.
    """

    with CursorContext(connection) as cursor:
        cursor.execute(query, queryargs)
        columns = tuple(map(itemgetter(0), cursor.description))
        types = tuple(mapmap(sqltypes_to_str, map(itemgetter(1), cursor.description)))

        with copen(path, "wt", encoding="utf-8", newline="") as csvfile:
            fw = csv.writer(csvfile)
            fw.writerow(columns)
            fw.writerow(types)
            progress = progress or Progress()

            for row in progress.track(iterfetch(cursor)):
                fw.writerow(row)


def import_csv_to_sqlite(connection: Connection, path: str, tablename: str, overwrite: bool = False) -> None:
    """Imports a csv file `path` into a SQLite database.
    If `overwrite` is True, the table will be dropped and recreated.
    """

    tablename = quote_identifier(tablename)

    with CursorContext(connection) as cursor:
        with copen(path, "rt", encoding="utf-8") as csvfile:
            fr = csv.reader(csvfile)
            columns = next(fr)
            types = next(fr)
            assert len(columns) == len(types)

            if overwrite:
                query = f"DROP TABLE IF EXISTS {tablename}"
                cursor.execute(query)

            zipped = zip(columns, mapmap(str_to_sqlitetype, types))
            typedcolsstr = ", ".join(quote_identifier(col) + " " + typ for col, typ in zipped)
            query = f"CREATE TABLE {tablename} ({typedcolsstr})"
            cursor.execute(query)

            valueparamsstr = ", ".join(repeat("?", len(columns)))
            query = f"INSERT INTO {tablename} VALUES ({valueparamsstr})"  # nosec
            cursor.executemany(query, fr)

    connection.commit()



"""This module is to avoid circular imports.
It should avoid any dependencies apart from the standard library.
"""

import os
import os.path
import re
from typing import Union

PathType = Union[str, os.PathLike]


class BaseDirEntry(os.PathLike):
    __slots__ = ("entry",)

    """ You cannot inherit from `os.DirEntry` (type 'nt.DirEntry' is not an acceptable base type),
        so use this wrapper class instead.
    """

    def __init__(self, entry: os.DirEntry) -> None:
        self.entry = entry

    @property
    def name(self) -> str:
        return self.entry.name

    @property
    def path(self) -> str:
        return self.entry.path

    def inode(self) -> int:
        return self.entry.inode()

    def is_dir(self, follow_symlinks=True) -> bool:
        return self.entry.is_dir(follow_symlinks=follow_symlinks)

    def is_file(self, follow_symlinks=True) -> bool:
        return self.entry.is_file(follow_symlinks=follow_symlinks)

    def is_symlink(self) -> bool:
        return self.entry.is_symlink()

    def stat(self, follow_symlinks=True) -> os.stat_result:
        return self.entry.stat(follow_symlinks=follow_symlinks)

    def __str__(self) -> str:
        return str(self.entry)

    def __repr__(self) -> str:
        return repr(self.entry)

    def __fspath__(self) -> str:
        return os.fspath(self.entry)


MyDirEntryT = Union[os.DirEntry, BaseDirEntry]


def entrysuffix(entry: MyDirEntryT) -> str:
    return os.path.splitext(entry.name)[1]


def to_dos_device_path(path: str) -> str:
    """Converts a standard absolute Windows path to a DOS device path aka local device paths.
    These paths are not normalized by the Win32 api and can be used to access otherwise unavailable files.
    """

    p = r"(?:(?:\\\\([^\\\/]+))|([a-z]:))([\\\/].*)"
    m = re.match(p, path, re.IGNORECASE)

    if m:
        server, drive, _path = m.groups()
        if drive is not None:
            return "\\\\?\\" + drive + _path
        elif server not in ("?", "."):
            return "\\\\?\\UNC\\" + server + _path

    return path


def to_dos_path(path: str) -> str:
    """Converts a DOS device path aka local device paths to a standard absolute Windows path.
    These paths have better compatibility and familiarity but also more limitations.
    """

    p = r"\\\\(?:\?|\.)\\(?:(unc\\)|([a-z]:))(.*)"
    m = re.match(p, path, re.IGNORECASE)
    if m:
        unc, drive, _path = m.groups()
        if unc:
            return "\\\\" + _path
        else:
            return drive + _path

    return path



import logging
import os
import os.path
from datetime import timedelta
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Dict, Optional

from .datetime import now
from .filesystem import mdatetime
from .object import args_to_key
from .time import MeasureTime

logger = logging.getLogger(__name__)


def cache(
    path: Path,
    duration: Optional[timedelta] = None,
    generator: bool = False,
    serializer: str = "pickle",
    serializer_kwargs: Optional[Dict[str, Any]] = None,
    deserializer_kwargs: Optional[Dict[str, Any]] = None,
    ignoreargs: bool = False,
    ignore_first_arg: bool = False,
    keyfunc: Optional[Callable[[tuple, dict], str]] = None,
    consume: bool = False,
    return_cached: bool = False,
    file_ext: Optional[str] = None,
    cached_only: bool = False,
) -> Callable[[Callable], Callable]:
    """Decorator to cache function calls. Doesn't take function arguments into regard.
    It's using `pickle` to deserialize the data. So don't use it with untrusted inputs.

    `path`: path to cache file or directory.
            This can be a template string where `ppv` is assigned the pickle protocol version.
    `duration`: maximum age of cache
    `generator`: set to True to store the results of generator objects
    `protocol`: pickle protocol version

    If `ignoreargs` is True, the cache won't take function arguments into regard.
            The path will be interpreted as template string to a file instead of a directory.
    """

    from . import pickle  # nosec

    serializer_kwargs = serializer_kwargs or {}
    deserializer_kwargs = deserializer_kwargs or {}

    if serializer == "pickle":
        from pickle import HIGHEST_PROTOCOL  # nosec

        _serializer_kwargs: Dict[str, Any] = {"protocol": HIGHEST_PROTOCOL, "safe": True}
        _deserializer_kwargs: Dict[str, Any] = {}
        write_iter = pickle.write_iter
        write_file = pickle.write_pickle
        read_iter = pickle.read_iter
        read_file = pickle.read_pickle
        key_to_hash = pickle.key_to_hash
        _hash_sep: Any = None

    elif serializer == "msgpack":
        from . import msgpack

        _serializer_kwargs = {"safe": True}
        _deserializer_kwargs = {}
        write_iter = msgpack.write_iter
        write_file = msgpack.write_msgpack
        read_iter = msgpack.read_iter
        read_file = msgpack.read_msgpack
        key_to_hash = msgpack.key_to_hash
        _hash_sep = None

    elif serializer == "json":
        from . import json

        _serializer_kwargs = {"ensure_ascii": False, "sort_keys": False, "default": None, "safe": True}
        _deserializer_kwargs = {"cls": None, "object_hook": None}
        write_iter = json.write_json_lines
        write_file = json.write_json
        read_iter = json.read_json_lines
        read_file = json.read_json
        key_to_hash = json.key_to_hash
        _hash_sep = {}

    else:
        raise ValueError(f"Invalid serializer: {serializer}")

    _serializer_kwargs.update(serializer_kwargs)
    _deserializer_kwargs.update(deserializer_kwargs)

    _pure_serializer_kwargs = _serializer_kwargs.copy()
    _pure_serializer_kwargs.pop("safe")

    if not generator and consume:
        raise ValueError("consume can only be used for generator")

    if duration is None:
        _duration = timedelta.max
    else:
        _duration = duration

    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def inner(*args: Any, **kwargs: Any) -> Any:
            strpath = os.fspath(path).format_map(_serializer_kwargs)

            if not ignoreargs:
                if ignore_first_arg:
                    args = args[1:]

                _file_ext = file_ext or ".p"

                if keyfunc is None:
                    hashstr = key_to_hash(args_to_key(args, kwargs, _hash_sep), **_pure_serializer_kwargs)
                else:
                    hashstr = keyfunc(args, kwargs)
                fullpath = os.path.join(strpath, hashstr + _file_ext)
            else:
                if keyfunc or file_ext:
                    raise ValueError("`keyfunc` or `file_ext` can only be specified if ignoreargs is False")
                if args or kwargs:
                    logger.warning("cache file decorator for %s called with arguments", func.__name__)
                fullpath = strpath

            try:
                invalid = now() - mdatetime(fullpath) > _duration
            except FileNotFoundError:
                invalid = True

            if invalid:
                if cached_only:
                    raise LookupError("Not in cache")

                cached = False
                if not ignoreargs:
                    path.mkdir(parents=True, exist_ok=True)

                if generator and not consume:
                    it = func(*args, **kwargs)
                    logger.debug("Writing iterable to <%s>", fullpath)
                    result: Any = write_iter(it, fullpath, **_serializer_kwargs)
                else:
                    with MeasureTime() as m:
                        if generator and consume:
                            result = list(func(*args, **kwargs))
                        else:
                            result = func(*args, **kwargs)
                    logger.debug("Result calculated in %s seconds and written to <%s>", m.get(), fullpath)
                    write_file(result, fullpath, **_serializer_kwargs)
            else:
                cached = True
                if generator and not consume:
                    logger.debug("Loading iterable from <%s>", fullpath)
                    result = read_iter(fullpath, **_deserializer_kwargs)
                else:
                    with MeasureTime() as m:
                        result = read_file(fullpath, **_deserializer_kwargs)
                    logger.debug("Result loaded from <%s> in %s seconds", fullpath, m.get())

            if return_cached:
                return cached, result
            else:
                return result

        return inner

    return decorator



import tracemalloc
from typing import Optional

from typing_extensions import Self


class MeasureMemory:
    __slots__ = ("total", "snapshot")

    total: Optional[int]

    def __init__(self) -> None:
        tracemalloc.start()

        self.total = None

    def _comp(self) -> int:
        snapshot_now = tracemalloc.take_snapshot()
        stats = snapshot_now.compare_to(self.snapshot, "lineno")
        return sum(stat.size for stat in stats)

    def __enter__(self) -> Self:
        self.snapshot = tracemalloc.take_snapshot()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.total = self._comp()

    def get(self) -> int:
        if self.total:
            return self.total
        else:
            return self._comp()

    def print(self, name: str) -> None:
        if self.total:
            total = self.total
        else:
            total = self._comp()

        print(f"{name} uses {total / 1024 / 1024:.3f} MiB of memory")



from collections.abc import Iterable, Mapping, MutableMapping
from typing import Any, Dict, Hashable
from typing import Iterable as IterableT
from typing import Iterator, List, Optional, Tuple, Union


class SequenceTree(MutableMapping):
    """SequenceTree is a tree class to store sequences, often called a Trie.
    It implements the MutableMapping interface, so it can be used like a dictionary.
    Compared to a dict, it offers a few advantages however:
    - common prefixes are only stored as one node
    - it supports the `longest_prefix` method
    """

    """ SequenceTree structure is not recursive, dicts are recursive.
		which way would be better? if SeqTrees were recursive, methods like values()
		would be easier to implement, but it would also incur another layer of
		abstraction which would make things slower?
	"""

    # __slots__ = ("root", "endkey")

    def __init__(self, data: Optional[Union[Mapping, Iterable]] = None, endkey: Hashable = "_") -> None:
        self.root: Dict[Hashable, Any] = {}
        self.endkey = endkey
        if data is None:
            pass
        elif isinstance(data, Mapping):  # mapping should be before Iterable
            for k, v in data.items():
                self[k] = v
        elif isinstance(data, Iterable):
            for k, v in data:
                self[k] = v
        else:
            raise ValueError("data should be a iterable or mapping")

    def iter_node(self, node) -> Iterator[Tuple[List[Any], Any]]:
        for k, v in node.items():
            if k == self.endkey:
                yield [], v
            else:
                for a, b in self.iter_node(v):
                    yield [k] + a, b

    @classmethod
    def fromtree(cls, tree: dict, endkey: Hashable = "_") -> "SequenceTree":
        seqtree = cls(endkey=endkey)
        seqtree.root = tree
        return seqtree

    def __contains__(self, keys: IterableT[Hashable]) -> bool:
        try:
            self[keys]
            return True
        except KeyError:
            return False

    def __getitem__(self, keys: IterableT[Hashable]) -> Any:
        node = self.root
        for word in keys:
            node = node[word]
        return node[self.endkey]

    def __setitem__(self, keys: IterableT[Hashable], value: Any) -> None:
        node = self.root
        for key in keys:
            node = node.setdefault(key, {})

        node[self.endkey] = value

    def set(self, keys: IterableT[Hashable], value: Any) -> bool:
        """Same as `__setitem__`, except it returns `True` if the key wasn't used before."""

        node = self.root
        for word in keys:
            node = node.setdefault(word, {})

        if self.endkey in node:
            node[self.endkey] = value
            return False
        else:
            node[self.endkey] = value
            return True

    def __delitem__(self, keys: IterableT[Hashable]) -> None:
        node = self.root
        for key in keys:
            node = node[key]
        del node[self.endkey]

    def _copy(self, node: dict) -> dict:
        ret = {}

        for k, v in node.items():
            if k != self.endkey:
                ret[k] = self._copy(v)
            else:
                ret[k] = v

        return ret

    def copy(self) -> "SequenceTree":
        """shallow copy"""

        tree = self._copy(self.root)
        return self.fromtree(tree, self.endkey)

    def __iter__(self) -> Iterator:
        return iter(self.keys())

    def __eq__(self, rhs: object) -> bool:
        if isinstance(rhs, SequenceTree):
            return self.root == rhs.root and self.endkey == rhs.endkey
        else:
            return False

    def __ne__(self, rhs: object) -> bool:
        if isinstance(rhs, SequenceTree):
            return self.root != rhs.root or self.endkey != rhs.endkey
        else:
            return False

    def __str__(self) -> str:
        return str(self.root)

    def __len__(self):
        raise NotImplementedError("use slower SequenceTree.calc_leaves()")

    def pop(self, keys: IterableT[Hashable]) -> Any:
        node = self.root
        for key in keys:
            node = node[key]
        return node.pop(self.endkey)

    def popdefault(self, keys: IterableT[Hashable], default: Optional[Any] = None) -> Any:
        try:
            return self.pop(keys)
        except KeyError:
            return default

    """ slower, I think?
	def popitem(self):
		try:
			keys = next(self.keys())
			return keys, self.pop(keys)
		except StopIteration:
			raise KeyError("tree is empty") from None
	"""

    def _popitem(self, node: Dict[Hashable, Any], branch: Tuple[Hashable, ...]) -> Tuple[Tuple[Hashable, ...], Any]:
        for key, value in node.items():
            if key != self.endkey:
                try:
                    ret = self._popitem(value, branch + (key,))
                    break
                except KeyError:
                    pass
            else:
                del node[key]
                ret = branch, value
                break
        else:
            raise KeyError("tree is empty") from None
        return ret

    def popitem(self) -> Tuple[Tuple[Hashable, ...], Any]:
        """leaves empty dicts behind. thus it's not good for looping."""

        return self._popitem(self.root, ())

    def _optimized(self, node: Dict[Hashable, Any]) -> Dict[Hashable, Any]:
        opt = {}

        for k, v in node.items():
            if k != self.endkey:
                opt[k] = self._optimized(v)
            else:
                opt[k] = v

        return {k: v for k, v in opt.items() if v}

    def optimized(self) -> "SequenceTree":
        """Returns shallow copy with empty branches removed."""

        tree = self._optimized(self.root)
        return SequenceTree.fromtree(tree, self.endkey)

    def update(self, other):
        raise NotImplementedError("update does not work yet")

    def clear(self) -> None:
        """Removes all leaves and branches from tree."""

        self.root.clear()

    def get(self, keys: IterableT[Hashable], default: Optional[Any] = None) -> Any:
        try:
            return self[keys]
        except KeyError:
            return default

    def setdefault(self, keys: IterableT[Hashable], default: Optional[Any] = None) -> Any:
        node = self.root
        for word in keys:
            node = node.setdefault(word, {})

        return node.setdefault(self.endkey, default)

    def _values(self, node: Dict[Hashable, Any]) -> Iterator[Any]:
        for key, value in node.items():
            if key != self.endkey:
                yield from self._values(value)
            else:
                yield value

    def values(self) -> Iterator[Any]:  # itervalues
        return self._values(self.root)

    def _keys(self, node: Dict[Hashable, Any], branch: Tuple[Hashable, ...]) -> Iterator[Tuple[Hashable, ...]]:
        for key, value in node.items():
            if key != self.endkey:
                yield from self._keys(value, branch + (key,))
            else:
                yield branch

    def keys(self) -> Iterator[Tuple[Hashable, ...]]:  # iterkeys
        return self._keys(self.root, ())

    def _items(
        self, node: Dict[Hashable, Any], branch: Tuple[Hashable, ...]
    ) -> Iterator[Tuple[Tuple[Hashable, ...], Any]]:
        for key, value in node.items():
            if key != self.endkey:
                yield from self._items(value, branch + (key,))
            else:
                yield branch, value

    def items(self) -> Iterator[Tuple[tuple, Any]]:  # iteritems
        return self._items(self.root, ())

    # others

    def longest_prefix(self, keys: Iterable, full: bool = True) -> Tuple[list, Any]:
        node = self.root
        ret = []
        for word in keys:
            try:
                node = node[word]
                ret.append(word)
            except KeyError:
                break
        if full:
            node = node[self.endkey]  # should this be within try-except like above? nope.
        return ret, node

    def get_node(self, keys):
        node = self.root
        for word in keys:
            node = node[word]

        return node

    @classmethod
    def _calc_max_depth(cls, node: dict) -> int:
        if node:
            return max(cls._calc_max_depth(v) for v in node.values()) + 1
        else:
            return 0

    def calc_max_depth(self) -> int:
        if self.root:
            return self._calc_max_depth(self.root) - 1
        else:
            return 0

    def _calc_branches(self, node: dict) -> int:
        ret = 0

        for k, v in node.items():
            if k != self.endkey and v:
                ret += self._calc_branches(v)
            else:
                ret += 1
        return ret

    def calc_branches(self) -> int:
        return self._calc_branches(self.root)

    def _calc_leaves(self, node: dict) -> int:
        ret = 0

        for k, v in node.items():
            if k != self.endkey:
                ret += self._calc_leaves(v)
            else:
                ret += 1
        return ret

    def calc_leaves(self) -> int:
        return self._calc_leaves(self.root)



class RingList:
    """bad complexity, moves all data if ring is full"""

    def __init__(self, length):
        self.data = []
        self.full = False
        self.max = length
        self.cur = 0

    def append(self, x):
        if self.full:
            for i in range(self.cur - 1):
                self.data[i] = self.data[i + 1]
            self.data[self.cur - 1] = x
        else:
            self.data.append(x)
            self.cur += 1
            if self.cur == self.max:
                self.full = True

    def get(self):
        return self.data

    def remove(self):
        if self.cur > 0:
            del self.data[self.cur - 1]
            self.cur -= 1

    def size(self):
        return self.cur

    def maxsize(self):
        return self.max

    def __str__(self):
        return "".join(self.data)



import logging
import os.path
from datetime import datetime, timedelta
from functools import partial, reduce, wraps
from sys import stdout
from time import sleep
from typing import Any, Callable, Iterable, Iterator, Optional, Sequence, TextIO, TypeVar, Union

from ._func import rename, renameobj  # noqa: F401
from .iter import nonstriter, retrier
from .typing import ExceptionsType

T = TypeVar("T")
U = TypeVar("U")
It = TypeVar("It", bound=Iterable)

logger = logging.getLogger(__name__)


class NotRetried(RuntimeError):
    pass


def identity(x: T) -> T:
    """Identity function."""

    return x


def nop() -> None:
    """Function which does absolutely nothing (aka pass, noop)."""

    pass


def partial_decorator(*args: Any, **kwargs: Any) -> Callable:
    """Same as `functools.partial` but applied as a decorator."""

    def decorator(func):
        return partial(func, *args, **kwargs)

    return decorator


def compose_two(f: Callable[[Any], Any], g: Callable[[Any], Any]) -> Callable[[Any], Any]:
    """compose_two(f, g) -> lambda x: f(g(x))"""

    return lambda x: f(g(x))


def compose(*functions: Callable[[Any], Any]) -> Callable[[Any], Any]:
    """compose(f, g, h) -> lambda x: f(g(h(x))).
    see: Function composition
    """

    return reduce(compose_two, functions, identity)


def apply(f: Callable[[T], U], x: T) -> U:  # *args, **kwargs
    """Same as `f(x)`."""

    return f(x)


def zipmap(funcs: Iterable[Callable[[T], U]], vals: Iterable[T]) -> Iterator[U]:
    """Applies a list of functions to a list of values."""

    return map(apply, funcs, vals)


def multiapply(funcs: Iterable[Callable], elm: Any) -> Any:
    """Applies functions `funcs` to element `elm` iteratively."""

    for func in funcs:
        elm = func(elm)

    return elm


def multimap(funcs: Sequence[Callable], it: Iterable) -> Iterator:
    """Applies functions `funcs` to each element of `it` iteratively."""

    for i in it:
        yield multiapply(funcs, i)


def deepmap(func: Callable, *iterables: Iterable) -> Iterator:
    """Simlar to `map`, but it maps recursively over sequences of sequnces.
    Returns a generator of generators.
    To output plain objects use `recmap` instead.

    Example:
    deepmap(lambda x: x*2, [1, [2, 3]]) -> (2, (4, 6))
    """

    if not iterables:
        raise ValueError("iterables must have a least one element")

    for args in zip(*iterables):
        try:
            its = tuple(map(nonstriter, args))
        except TypeError:
            yield func(*args)
        else:
            yield deepmap(func, *its)


def outermap(func: Callable[[T], U], iterable: It) -> Union[It, U]:
    """Examples:
    outermap(list, (1, (2, 3))) -> [1, [2, 3]]
    outermap(sum, (1, (2, 3)) -> 6
    """

    try:
        it = nonstriter(iterable)
    except TypeError:
        return iterable
    else:
        return func(outermap(func, i) for i in it)


def recmap(func: Callable, iterable: Iterable[Any]) -> list:
    """Simlar to `map`, but it maps recursively over sequences of sequnces and puts the result into a list of lists.
    To return a generator of generators use `deepmap` instead.

    Example:
    recmap(lambda x: x*2, [1, [2, 3]]) -> [2, [4, 6]]
    """

    return outermap(list, deepmap(func, iterable))


def call_repeated(num: int) -> Callable[[Callable], Callable]:
    """Function decorator to call decorated function `num` times with the same arguments.
    Returns the results of the last call.
    """

    if num < 1:
        raise ValueError("num must be larger than 0")

    def dec(func):
        @wraps(func)
        def inner(*args, **kwargs):
            last = None
            for _i in range(num):
                last = func(*args, **kwargs)
            return last

        return inner

    return dec


def print_return_type(func: Callable, file: TextIO = stdout) -> Callable:
    """Wraps function to print the return type after calling."""

    @wraps(func)
    def inner(*args, **kwargs):
        ret = func(*args, **kwargs)
        print(type(ret), file=file)
        return ret

    return inner


def get_callable_name(func: Callable) -> str:
    """Unwraps common function wrappers and tries to return the full qualified function name."""

    if isinstance(func, partial):
        func = func.func

    return getattr(func, "__qualname__", repr(func))


def retry(
    func: Callable[[], T],
    waittime: float,
    exceptions: ExceptionsType = Exception,
    attempts: int = -1,
    multiplier: float = 1,
    jitter: float = 0,
    max_wait: Optional[float] = None,
    jitter_dist: str = "uniform",
    waitfunc: Callable[[float], Any] = sleep,
) -> T:
    """Retry function `func` multiple times in case of raised `exceptions`.
    See `genutility.iter.retrier()` for the remaining arguments.
    Reraises the last exception in case the function call doesn't succeed after retrying.
    """

    last_exception: Optional[Exception] = None
    for i in retrier(waittime, attempts, multiplier, jitter, max_wait, jitter_dist, waitfunc):
        try:
            return func()
        except exceptions as e:
            name = get_callable_name(func)
            logger.info("Attempt %s (%s) failed: %s", i + 1, name, e)
            last_exception = e

    if last_exception:
        raise last_exception  # pylint: disable=raising-bad-type
    else:
        raise NotRetried


def default_except(
    exceptions: ExceptionsType, default: T, func: Callable[..., U], *args: Any, **kwargs: Any
) -> Union[T, U]:
    """Call `func(*args, **kwargs)` and turn `exceptions` into `default`."""

    try:
        return func(*args, **kwargs)
    except exceptions:
        return default


class CustomCache:
    """Class to build decorator cache function using custom reader and writer functions.
    The cache method ignores all arguments of the decorated function.

    Example:
    ```
    cc = CustomCache(read_pickle, write_pickle)

    @cc.cache("func-cache.p")
    # arg is ignored, so the cache will return the result of the argument which was supplied
    # when the cache file was created.
    def func(arg):
            return arg
    ```
    a = func(1) # cache created, a == 1
    b = func(2) # cache loaded, b == 1
    """

    def __init__(self, reader: Callable, writer: Callable) -> None:
        self.reader = reader
        self.writer = writer

    def cache(self, path: str) -> Callable:
        def dec(func: Callable) -> Callable:
            @wraps(func)
            def inner(*args: Any, **kwargs: Any) -> Any:
                if os.path.exists(path):
                    logger.debug("Loading object from cache %s", path)
                    return self.reader(path)
                else:
                    logger.debug("Saving obect to cache %s", path)
                    result = func(*args, **kwargs)
                    self.writer(result, path)
                    return result

            return inner

        return dec


class RunScheduled:
    def __init__(self, delta: timedelta, func: Callable):
        self.delta = delta
        self.func = func
        self.lastrun: Optional[datetime] = None

    def __call__(self, *args, **kwargs) -> None:
        now = datetime.now()
        if self.lastrun is None or now - self.lastrun > self.delta:
            self.func(*args, **kwargs)
            self.lastrun = now


def applymap(func: Callable[..., T], it: Iterable[tuple]) -> Iterator[T]:
    """Maps `func` over the unpacked `it`."""

    return (func(*args) for args in it)



from typing import Any, Callable, Iterable, List, Optional, Tuple, TypeVar

from pandas import DataFrame, Series

from .func import identity

T = TypeVar("T")


def intersect_columns(*dfs: DataFrame, sort=False) -> Tuple[DataFrame, ...]:
    """Takes multiple dataframes and removes the columns which are not present in the other dataframes.
    The order of columns will be equalized.
    """

    cols = tuple(set(df.columns) for df in dfs)
    goodset = set.intersection(*cols)
    if sort:
        good = sorted(goodset)
    else:
        good = list(goodset)
    dfs = tuple(df[good] for df in dfs)
    assert len(set(map(lambda df: tuple(df.columns), dfs))) == 1

    return dfs


def pandas_json(obj: Any) -> Any:
    """Can be used for the `json.dump` `default` argument
    to make some pandas objects JSON serializable.
    """

    if isinstance(obj, DataFrame):
        return dict(obj)
    elif isinstance(obj, Series):
        return tuple(obj)

    raise TypeError(f"object of type {type(obj)} cannot be JSON serialized: {obj}")


def strlist(sep: str) -> Callable[[str], List[str]]:
    def inner(s):
        if s:
            return s.split(sep)
        else:
            return []

    return inner


def dataframe_to_dict(
    df: DataFrame,
    empty: bool = False,
    cellfunc: Optional[Callable] = None,
    dictcls: Callable[[Iterable[Tuple[str, Any]]], T] = dict,
) -> T:
    """Turns dataframes into trees. Columns first, then rows"""

    if cellfunc is None:
        cellfunc = identity

    def cols(series):
        for rowname, rowdata in series.iteritems():
            value = cellfunc(rowdata)
            if value or empty:
                yield rowname, value

    def rows(df):
        for colname, coldata in df.iteritems():
            value = dictcls(cols(coldata))
            if value or empty:
                yield colname, value

    return dictcls(rows(df))



import logging
from typing import Collection, Dict, List, Optional, Sequence

import wx
from typing_extensions import Self
from wx.lib.mixins.listctrl import ColumnSorterMixin

logger = logging.getLogger(__name__)


class AdvancedListCtrl(wx.ListCtrl, ColumnSorterMixin):
    # wx.LC_SORT_* breaks idtopos
    def __init__(self, *args, **kwargs):
        wx.ListCtrl.__init__(self, *args, **kwargs)
        self.columns = 0
        self._init()
        ColumnSorterMixin.__init__(self, 0)

        # self.Bind(wx.EVT_LIST_DELETE_ITEM, self.OnDeleteItem)
        # self.Bind(wx.EVT_LIST_INSERT_ITEM, self.OnInsertItem)

    def _init(self):
        self.rows = 0
        self.itemDataMap = {}
        self.insertcount = 0
        self.idtopos = {}

    def GetListCtrl(self) -> Self:
        return self

    def ClearAll(self, clear_columns: bool = True) -> None:
        if clear_columns:
            wx.ListCtrl.ClearAll(self)
            self.columns = 0
        else:
            wx.ListCtrl.DeleteAllItems(self)
        self._init()

    def Setup(self, columns: Collection[str]) -> None:
        self.columns = len(columns)
        self.SetColumnCount(self.columns)
        for i, col in enumerate(columns):
            self.InsertColumn(i, col)

    def Append(self, items: Sequence[str]) -> None:
        if len(items) != self.columns:
            raise ValueError("Length of items doesn't match available columns")

        pos = self.InsertItem(self.rows, items[0])
        assert pos == self.rows
        # logger.debug("Appending Item on rows->pos: {}->{}".format(self.rows, pos))

        self.rows += 1
        self.SetItemData(pos, self.insertcount)
        self.itemDataMap[self.insertcount] = items

        """for id, pos in self.idtopos.items():
            if pos >= event.Index:
                self.idtopos[id] = pos + 1"""

        self.idtopos[self.insertcount] = pos
        self.insertcount += 1

        for col in range(1, self.columns):
            self.SetItem(pos, col, items[col])

        return self.insertcount - 1

    def Edit(self, pos: int, items: Dict[int, str]):
        """items is dict with col as key and string als value"""

        for col, val in items.items():
            self.SetItem(pos, col, val)

    def GetSelections(self) -> List[int]:
        indexes = []
        pos = self.GetFirstSelected()
        while pos != -1:
            indexes.append(pos)
            pos = self.GetNextSelected(pos)
        return indexes

    def DeleteSelectedItems(self) -> List[int]:
        # might be faster with autoarrange (listbox only??) off, no reversed -> autoarrange on
        indexes = self.GetSelections()

        for index in reversed(indexes):
            self.DeleteItem(index)
            self.rows -= 1
            id = self.GetItemData(index)
            del self.itemDataMap[id]
            del self.idtopos[id]

        return indexes

    def Get(self, pos: int = 0) -> str:
        if pos < 0:
            return self.GetItem(self.GetItemCount() + 1 + pos).GetText()
        else:
            return self.GetItem(pos).GetText()

    def IDtoPos(self, id: int) -> int:
        return self.idtopos[id]

    def Delete(self, pos: int = 0) -> Optional[str]:
        if pos < 0:
            index = self.GetItemCount() + 1 + pos
        else:
            index = pos
        if self.GetItemCount() > pos:
            item = self.GetItem(index)
            ret = item.GetText()
            id = self.GetItemData(index)
            del self.itemDataMap[id]
            del self.idtopos[id]

            self.rows -= 1
            self.DeleteItem(index)
            return ret
        else:
            return None


class AdvancedItemContainerMixin:
    def DeleteSelectedItems(self) -> List[int]:
        # might be faster with autoarrange of, no reversed -> autoarrange on
        indexes = self.GetSelections()

        for index in reversed(indexes):
            self.Delete(index)

        return indexes


class AdvancedListBox(wx.ListBox, AdvancedItemContainerMixin):
    def __init__(self, *args, **kwargs) -> None:
        wx.ListBox.__init__(self, *args, **kwargs)



from typing import Optional

import numpy as np

from .callbacks import Progress


class LinearRegression:
    """Linear regression optimized by gradient descent."""

    def __init__(self, X: np.ndarray, y: np.ndarray, alpha: float = 0.01) -> None:
        """
        X: float[A, B]
        Y: float[A]
        """

        if len(X.shape) != 2:
            raise ValueError("X must be 2-dimensional")
        if len(y.shape) != 1:
            raise ValueError("y must be 1-dimensional")

        self.K = X.shape[1] + 1  # one extra dimension for the bias term
        self.M = len(y)
        self.y = y
        self.α = alpha

        row = np.ones((self.M, 1))
        self.X = np.concatenate([row, X], axis=-1)
        assert self.X.shape == (self.M, self.K)

        self.weights = np.random.random_sample(self.K)

    def predict(self, x: np.ndarray) -> float:
        """
        x: float[B]
        """

        return np.matmul(x, self.weights)

    def epoch(self) -> None:
        self.weights -= self.α * np.matmul(self.X.T, self.predict(self.X) - self.y)  # [K]

    def fit(self, n_iter: int = 100, progress: Optional[Progress] = None) -> None:
        progress = progress or Progress()
        for _i in progress.track(range(n_iter), description="SGD itertion"):
            self.epoch()

    def getParams(self) -> np.ndarray:
        """
        returns: float[B-1]
        """

        return self.weights[1:]

    def getIntercept(self) -> float:
        return self.weights[0]



import errno
import gzip
import json
import logging
import os
import os.path
import socket
import ssl
from email.utils import parsedate_to_datetime
from typing import IO, TYPE_CHECKING, Any, Callable, Dict, List, Optional, Tuple, Union
from urllib import request
from urllib.error import URLError

from typing_extensions import Self

from .exceptions import DownloadFailed
from .file import Tell, copyfilelike
from .filesystem import safe_filename
from .iter import first_not_none
from .url import get_filename_from_url

if TYPE_CHECKING:
    from email.message import Message
    from http.client import HTTPMessage
    from http.cookiejar import CookieJar

JsonObject = Union[List[Any], Dict[str, Any]]

logger = logging.getLogger(__name__)
DEFAULT_TIMEOUT = 120.0


def install_ceritfi_opener() -> None:
    import certifi

    logger.warning("Using certifi store: %s", certifi.where())

    context = ssl.SSLContext()
    context.verify_mode = ssl.CERT_REQUIRED
    context.load_verify_locations(certifi.where())
    context.check_hostname = True
    https_handler = request.HTTPSHandler(context=context)
    opener = request.build_opener(https_handler)
    request.install_opener(opener)


class HTTPError(Exception):
    def __init__(self, *args, **kwargs):
        response = kwargs.pop("response", None)
        Exception.__init__(self, *args, **kwargs)
        self.response = response


class ContentInvalidLength(HTTPError, DownloadFailed):
    def __init__(self, path, expected, received):
        HTTPError.__init__(self, path, expected, received)


class TimeOut(HTTPError, DownloadFailed):
    pass


class DownloadInterrupted(HTTPError, DownloadFailed):
    pass


class NoRedirect(HTTPError):
    pass


def parsedate_to_timestamp(datestr: str) -> float:
    return parsedate_to_datetime(datestr).timestamp()


def get_filename(headers: "Message") -> Optional[str]:
    return headers.get_filename()


class URLRequestBuilder:
    def __init__(
        self, cookiejar: Optional["CookieJar"] = None, basicauth: Optional[Tuple[str, str, str]] = None
    ) -> None:
        handlers: List[request.BaseHandler] = []

        if cookiejar:
            handlers.append(request.HTTPCookieProcessor(cookiejar))
        if basicauth:
            top_level_url, username, password = basicauth
            password_mgr = request.HTTPPasswordMgrWithDefaultRealm()
            password_mgr.add_password(None, top_level_url, username, password)
            handlers.append(request.HTTPBasicAuthHandler(password_mgr))

        if handlers:
            opener = request.build_opener(*handlers)
            self.openfunc: Callable = opener.open
        else:
            self.openfunc = request.urlopen

    def getfunc(self):
        return self.openfunc

    def request(
        self,
        url: str,
        headers: Optional[dict] = None,
        timeout: float = DEFAULT_TIMEOUT,
        context: Optional[ssl.SSLContext] = None,
    ) -> "URLRequest":
        return URLRequest(url, headers, timeout, context, openfunc=self.openfunc)


class FileLike:
    def __init__(self, urlrequest):
        self.urlrequest = urlrequest

        self.content_length = self.urlrequest._content_length()

        self.response = Tell(self.urlrequest.response)
        if self.urlrequest.headers.get("Content-Encoding", None) == "gzip":
            # if that does not work, use BytesIO
            logger.info("Using gzip")
            self.fp = gzip.GzipFile(fileobj=self.response)
        else:
            self.fp = self.response

    def close(self) -> None:
        self.fp.close()

        transferred = self.response.tell()  # is this correct?
        if self.content_length is not None and self.content_length != transferred:
            raise ContentInvalidLength(None, self.content_length, transferred)

    def __enter__(self) -> IO[bytes]:
        return self.fp

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.close()


def get_redirect_url(url, headers=None):
    import requests

    r = requests.get(url, headers=headers, timeout=DEFAULT_TIMEOUT, allow_redirects=False)
    r.raise_for_status()

    try:
        if r.status_code in {301, 302, 303, 307, 308}:
            return r.headers["Location"]
        else:
            raise NoRedirect(f"Unexpected status code: {r.status_code}", response=r)

    except KeyError:  # Location not provided. is this really raised?
        raise HTTPError("Location header not found", response=r)


class URLRequest:
    headers: "HTTPMessage"

    def __init__(
        self,
        url: str,
        headers: Optional[dict] = None,
        timeout: float = DEFAULT_TIMEOUT,
        context: Optional[ssl.SSLContext] = None,
        cookiejar: Optional["CookieJar"] = None,
        basicauth: Optional[Tuple[str, str, str]] = None,
        openfunc: Optional[Callable] = None,
    ) -> None:
        self.url = url
        self.timeout = timeout
        headers = headers or {}

        if not openfunc:
            openfunc = URLRequestBuilder(cookiejar, basicauth).getfunc()

        req = request.Request(url, data=None, headers=headers)
        self.response = openfunc(req, timeout=timeout, context=context)
        self.headers = self.response.info()

    def _close(self):  # unused
        self.response.close()

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.response.close()

    def _content_length(self) -> Optional[int]:
        try:
            content_length = self.headers["Content-Length"]
            return int(content_length)
        except (KeyError, ValueError, TypeError):
            return None

    def _last_modified(self) -> Optional[float]:
        try:
            last_modified = self.headers["Last-Modified"]
            return parsedate_to_timestamp(last_modified)
        except (KeyError, TypeError, ValueError):
            return None

    def get_redirect_url(self) -> str:
        self.response.close()
        return self.response.geturl()

    def _load(self) -> bytes:
        with FileLike(self) as fp:
            try:
                return fp.read()
            except URLError:
                raise TimeOut(f"Timed out after {self.timeout}s", response=self.response)

    def _json(self) -> JsonObject:
        with FileLike(self) as fp:
            try:
                return json.load(fp)
            except URLError:
                raise TimeOut(f"Timed out after {self.timeout}s", response=self.response)

    def _download(
        self,
        basepath: str,
        filename: Optional[str] = None,
        fn_prio: Optional[Tuple[int, int, int, int]] = None,
        overwrite: bool = False,
        suffix: str = ".partial",
        report: Optional[Callable[[int, int], None]] = None,
    ) -> Tuple[Optional[int], str]:
        """PermissionError: [WinError 32] Der Prozess kann nicht auf die Datei zugreifen,
        da sie von einem anderen Prozess verwendet wird:
        'G:\\PUBLIC\\Audio\\Podcasts\\Radio Nukular\\radio_nukular_057.m4a.partial'
        -> 'G:\\PUBLIC\\Audio\\Podcasts\\Radio Nukular\\radio_nukular_057.m4a'
        """

        logger.debug("Downloading %s to %s", self.url, basepath)

        if fn_prio is None:
            fn_prio = (0, 1, 2, 3)

        filenames_ = {
            0: filename,
            1: get_filename(self.headers),
            2: get_filename_from_url(self.response.geturl()),  # url after redirect
            3: get_filename_from_url(self.url),
        }

        filenames = [filenames_[p] for p in fn_prio]
        logger.info("Filenames: {}".format(", ".join(map(str, filenames))))

        filename = first_not_none(filenames)

        if not filename:
            raise ValueError("Please provide a filename")

        filename = safe_filename(filename)
        fullpath = os.path.join(basepath, filename)
        tmppath = fullpath + suffix

        if not overwrite and os.path.exists(fullpath):
            raise FileExistsError(errno.EEXIST, "File already exists", fullpath)

        os.makedirs(basepath, exist_ok=True)

        content_length = self._content_length()

        try:
            # actually download the file
            with open(tmppath, "wb") as out:
                # https://www.ietf.org/mail-archive/web/httpbisa/current/msg27484.html
                transferred = copyfilelike(self.response, out, content_length, report=report)

        except (socket.timeout, URLError):
            logger.warning(f"Timeout after {self.timeout}s at {self.response.geturl()}: {self.headers}")
            raise TimeOut(f"Timed out after {self.timeout}s", response=self.response)

        except ConnectionResetError as e:
            logger.warning("Connection was reset during download: %s", str(e))
            raise DownloadInterrupted("Connection was reset during download")

        except FileNotFoundError:
            # can be raised on windows for example when the file path is too long
            raise

        except Exception:
            logger.exception("Downloading failed mid file. Handle.")
            raise

        if content_length and content_length != transferred:
            logger.info(f"{content_length} {transferred}")
            raise ContentInvalidLength(tmppath, content_length, transferred)

        last_modified = self._last_modified()
        if last_modified:
            os.utime(tmppath, (-1, last_modified))

        os.rename(tmppath, fullpath)

        return (content_length, filename)

    def load(self) -> bytes:
        try:
            return self._load()
        finally:
            self.response.close()

    def json(self) -> JsonObject:
        try:
            return self._json()
        finally:
            self.response.close()

    def download(
        self,
        basepath: str,
        filename: Optional[str] = None,
        fn_prio: Optional[Tuple[int, int, int, int]] = None,
        overwrite: bool = False,
        suffix: str = ".partial",
        report: Optional[Callable[[int, int], None]] = None,
    ) -> Tuple[Optional[int], str]:
        try:
            return self._download(basepath, filename, fn_prio, overwrite, suffix, report)
        finally:
            self.response.close()


if __name__ == "__main__":
    from argparse import ArgumentParser

    parser = ArgumentParser(description="Downloads url to file")
    parser.add_argument("url", help="url to download")
    parser.add_argument("dir", help="output directory")
    parser.add_argument("-f", "--filename", help="filename")
    args = parser.parse_args()

    URLRequest(args.url).download(args.dir, args.filename)



from logging import Formatter, LogRecord
from logging import basicConfig as _basicConfig
from logging import root as _root
from typing import Any, Dict, Literal, Optional, Type

from .datetime import datetime_from_utc_timestamp


def basicConfig(formatter: Optional[Formatter] = None, **kwargs) -> None:
    """Same as `logging.basicConfig`, but allows to specify a formatter instance
    to be added to the handlers instead of the default `logging.Formatter`
    """

    if formatter is not None and (kwargs.get("format") is not None or kwargs.get("datefmt") is not None):
        raise ValueError("Cannot specify `format` or `datefmt` with `formatter`")

    _basicConfig(**kwargs)

    if formatter is not None:
        for handler in _root.handlers:
            handler.setFormatter(formatter)


class IsoDatetimeFormatter(Formatter):
    """Displays the time in ISO 8601 format.
    Instead of passing a formatting string to `datefmt` in the initializer, `sep` and `timespec`
    (see `datetime.datetime.isoformat`) as well as `aslocal` (see genutility.datetime.datetime_from_utc_timestamp`)
    can be passed.
    """

    def __init__(
        self,
        fmt: Optional[str] = None,
        datefmt: Type[None] = None,
        style: Literal["%", "{", "$"] = "%",
        validate: bool = True,
        sep: str = "T",
        timespec: str = "auto",
        aslocal: bool = False,
    ) -> None:
        Formatter.__init__(self, fmt, datefmt, style, validate)
        assert datefmt is None
        self.sep = sep
        self.timespec = timespec
        self.aslocal = aslocal

    def formatTime(self, record: LogRecord, datefmt: Type[None]) -> str:
        return datetime_from_utc_timestamp(record.created, aslocal=self.aslocal).isoformat(self.sep, self.timespec)


class OverwriteFormatter(Formatter):
    """Usually the default Formatter disallows overwriting of built-in fields
    like filename, funcName, lineno and so on. Here you can specify a mapping
    to set these values anyway.

    Example:
            formatter = OverwriteFormatter({"func_name": "funcName"}, "%(funcName)s: %(message)s")
            and then pass `extra={"func_name": "thename"}` to the logging calls.
            If the the custom fields are not passed, the original values will be used.
    """

    def __init__(self, map: Dict[str, str], *args: Any, **kwargs: Any) -> None:
        self.map = map
        Formatter.__init__(self, *args, **kwargs)

    def format(self, record: LogRecord) -> str:
        for k, v in self.map.items():
            try:
                setattr(record, v, getattr(record, k))
            except AttributeError:
                pass

        return Formatter.format(self, record)


if __name__ == "__main__":
    import logging

    logger = logging.getLogger(__name__)
    handler = logging.StreamHandler()
    formatter = OverwriteFormatter({"func_name": "funcName"}, "%(funcName)s: %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    logger.warning("asd")
    logger.warning("asd", extra={"func_name": "thename"})



from typing import Iterable, Iterator, Sequence, Tuple, TypeVar

T = TypeVar("T")


def obj2tuple(obj: T) -> Tuple[T]:
    return (obj,)


def row_indices(n: int, square_length: int) -> Iterable[int]:
    """Returns the one-dimensional indices of all the cells belonging to the
    same row as index `n`.
    """

    j = n // square_length
    return range(j * square_length, (j + 1) * square_length)


def col_indices(n: int, square_length: int) -> Iterable[int]:
    """Returns the one-dimensional indices of all the cells belonging to the
    same column as index `n`.
    """

    j = n % square_length
    return range(j, square_length * (square_length - 1) + j + 1, square_length)


def subblock_indices(n: int, outer_square_length: int, inner_square_length: int) -> Iterator[int]:
    """Returns the one-dimensional indices of all the cells belonging to the
    same block as index `n`.
    `outer_square_length` is the number of cells per block.
    `inner_square_length` current has to be sqrt(block_size).
    """

    i = outer_square_length
    j = inner_square_length

    if n >= i * i:
        raise ValueError("n not contained in matrix")

    if j * j != i:
        raise ValueError("Currently only squares are supported")

    x, y = (n % i // j, n // i // j)

    for iy in range(j):
        for ix in range(j):
            yield (y * i * j) + (i * iy) + (x * j) + (ix)


def triangular_indices(n: int) -> Iterator[Tuple[int, int]]:
    """Returns all combinations of indices for a sequence of length `n` with duplicates and self-combinations.
    triangular_indices(4) -> (0, 1), (0, 2), (0, 3), (1, 2), (1, 3), (2, 3)
    """

    for i in range(n - 1):
        for j in range(i + 1, n):
            yield i, j


def to_2d_index(n: int, width: int) -> Tuple[int, int]:
    return n // width, n % width


def indices_2d(n: int, width: int) -> Iterator[Tuple[int, int]]:
    for i in range(n):
        yield i // width, i % width


def window_combinations_indices(size: int, window_size: int) -> Iterator[Tuple[int, int]]:
    """Yields all pairs of indices of a sequence of length `size` which are within a window
    of size `window_size`.

    Example:
            window_combinations_indices(3, 2) -> (0, 1), (1, 2)
            window_combinations_indices(4, 3) -> (0, 1), (0, 2), (1, 2), (1, 3), (2, 3)
    """

    if window_size > size:
        raise ValueError("window size cannot exceed size")

    for a in range(0, size - 1):
        end = min(a + window_size, size)
        for b in range(a + 1, end):
            yield a, b


def _combination_indices(start: int, sizes: Sequence[int]) -> Iterator[Tuple[int, ...]]:
    """earlier (left) dimensions can never be higher than later (right) dimensions."""

    if len(sizes) > 1:
        for i in range(start, sizes[0]):
            for j in _combination_indices(i, sizes[1:]):
                yield (i,) + j
    else:
        yield from map(obj2tuple, range(start, sizes[0]))


def combination_indices(*sizes: int) -> Iterator[Tuple[int, ...]]:
    """Given sequences of size a, b, c, ... yield the indices to index into each of them,
    so that all combinations of elements from the sequences are returned.
    """

    return _combination_indices(0, sizes)



import logging
import os.path
import sqlite3
import warnings
from collections import UserDict
from functools import lru_cache
from itertools import chain, repeat
from os import fspath
from pathlib import Path
from typing import Any, Callable, Collection, Dict, Iterable, Iterator, List, Optional, Sequence, Tuple, TypeVar

from tls_property import tls_property
from typing_extensions import Self

from .exceptions import NoResult
from .filesystem import EntryType, normalize_seps
from .sql import fetchone, iterfetch
from .sqlite import quote_identifier
from .typing import HashableContainer

logger = logging.getLogger(__name__)

T = TypeVar("T")


class Keys(UserDict):
    def __hash__(self):
        return hash(tuple(self))


class GenericDb:
    order_col = "_order"

    def __init__(self, dbpath: str, table: str, debug: bool = True, allow_add: bool = True) -> None:
        if sqlite3.sqlite_version_info < (3, 35, 0):
            warnings.warn(
                f"SQLite version 3.35.0 or higher required for some features (current version {sqlite3.sqlite_version}). ",
                stacklevel=2,
            )

        self.table = quote_identifier(table)
        self._primary = self.primary()
        self._auto = self.auto()
        self._mandatory = self.mandatory()
        self._derived = self.derived()

        self.dbpath = dbpath
        self.setup()

        if logger.isEnabledFor(logging.DEBUG):
            logger.info("SQLite tracing enabled")
            sqlite3.enable_callback_tracebacks(True)
            self.connection.set_trace_callback(self.trace)

        # verify columns
        sql = f"SELECT c.name FROM pragma_table_info({self.table}) c"  # nosec
        self.cursor.execute(sql)
        file_cols = {name for (name,) in self.cursor.fetchall()}
        db_cols = {n: t for n, t, v in chain(self._primary, self._auto, self._mandatory, self._derived)}

        if self.order_col in db_cols:
            raise ValueError(f"`{self.order_col}` is a reserved column name")

        missing_cols = db_cols.keys() - file_cols
        if missing_cols:
            if allow_add:
                logger.info("Adding columns to database: %s", ", ".join(missing_cols))
                for name in missing_cols:
                    sql = f"ALTER TABLE {self.table} ADD {name} {db_cols[name]}"
                    self.cursor.execute(sql)
                self.commit()
            else:
                raise ValueError(f"Database is missing columns: {', '.join(missing_cols)}")

        self._get_latest_sql = lru_cache(maxsize=128)(self._get_latest_sql)  # type: ignore[assignment,method-assign]

    @tls_property
    def connection(self):
        return sqlite3.connect(self.dbpath, detect_types=sqlite3.PARSE_DECLTYPES, check_same_thread=False)

    @tls_property
    def cursor(self):
        return self.connection.cursor()

    def trace(self, query: str) -> None:
        logger.debug("SQL trace: %s", query)

    @classmethod
    def latest_order_by(cls) -> Tuple[str, str, str]:
        raise NotImplementedError

    @classmethod
    def primary(cls) -> List[Tuple[str, str, str]]:
        """Never explicitly specified.
        Not needed if there is a PRIMARY KEY in one of the other fields.
        """

        raise NotImplementedError

    @classmethod
    def auto(cls) -> List[Tuple[str, str, str]]:
        """Explicitly specified value.
        Can never be a ? placeholder, but must be a SQL value like `datetime('now')`.
        """

        raise NotImplementedError

    @classmethod
    def mandatory(cls) -> List[Tuple[str, str, str]]:
        """Mandatory fields used to retrieve rows.
        Used for all search queries.
        """

        raise NotImplementedError

    @classmethod
    def derived(cls) -> List[Tuple[str, str, str]]:
        """Optional fields used to retrieve rows.
        Used for search queries when given.
        """

        return []

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.close()

    def __len__(self) -> int:
        sql = f"SELECT count(*) FROM {self.table}"  # nosec
        self.cursor.execute(sql)
        (result,) = fetchone(self.cursor)
        return result

    def __bool__(self) -> bool:
        sql = f"SELECT EXISTS (SELECT 1 FROM {self.table})"  # nosec
        self.cursor.execute(sql)
        (result,) = fetchone(self.cursor)
        return result == 1

    def close(self) -> None:
        """Only closes connections and cursors opened in the current thread"""

        self.cursor.close()
        self.connection.close()

    def setup(self) -> None:
        fields = ", ".join(f"{n} {t}" for n, t, v in chain(self._primary, self._auto, self._mandatory, self._derived))

        # unique = "UNIQUE({})".format(n for n, t, v in chain(self._mandatory, self._derived))
        # add to sql below?
        sql = f"CREATE TABLE IF NOT EXISTS {self.table} ({fields})"
        self.cursor.execute(sql)
        self.commit()

    def normalize_mandatory(self, mandatory: Sequence) -> Sequence:
        return mandatory

    def commit(self) -> None:
        self.connection.commit()

    def _args(
        self,
        mandatory: Sequence[Any],
        derived: Dict[str, Any],
        ignore_null: bool = True,
    ) -> tuple:
        if ignore_null:
            return tuple(
                chain(
                    self.normalize_mandatory(mandatory),
                    (derived[n] for n, t, v in self._derived if n in derived),
                )
            )
        else:
            return tuple(chain(self.normalize_mandatory(mandatory), (derived.get(n) for n, t, v in self._derived)))

    def _args_many(
        self,
        mandatory: Iterable[Sequence[Any]],
        derived: Optional[Iterable[Dict[str, Any]]] = None,
        ignore_null: bool = True,
    ) -> Iterator[tuple]:
        derived = derived or repeat({})
        for _mandatory, _derived in zip(mandatory, derived):
            yield self._args(_mandatory, _derived, ignore_null)

    def iter(
        self, only: HashableContainer[str] = frozenset(), no: HashableContainer[str] = frozenset()
    ) -> Iterator[tuple]:
        if only and no:
            raise ValueError("Only `only` or `no` can be specified")

        if only:
            fields = ", ".join(
                n for n, t, v in chain(self._primary, self._auto, self._mandatory, self._derived) if n in only
            )
        else:
            fields = ", ".join(
                n for n, t, v in chain(self._primary, self._auto, self._mandatory, self._derived) if n not in no
            )

        sql = f"SELECT {fields} FROM {self.table}"  # nosec
        self.cursor.execute(sql)
        return iterfetch(self.cursor)

    def _filtered_derived(self, derived: Collection[str], select: str) -> Iterable[Tuple[str, str, str]]:
        if select == "null":
            return ((n, t, v) for n, t, v in self._derived if n not in derived)
        elif select == "not-null":
            return ((n, t, v) for n, t, v in self._derived if n in derived)
        elif select == "all":
            return self._derived
        else:
            raise ValueError("select must be one of: null, not-null, all")

    def _get_fields(self, only: HashableContainer[str], no: HashableContainer[str]) -> List[str]:
        if only and no:
            raise ValueError("Only `only` or `no` can be specified")

        if only:
            fields = [n for n, t, v in chain(self._primary, self._auto, self._mandatory, self._derived) if n in only]
        else:
            fields = [n for n, t, v in chain(self._primary, self._auto, self._mandatory, self._derived) if n not in no]

        if not fields:
            raise ValueError("No output fields selected")

        return fields

    def _get_latest_sql(
        self, derived_keys: Collection[str], ignore_null: bool, only: HashableContainer[str], no: HashableContainer[str]
    ) -> str:
        fields = ", ".join(self._get_fields(only, no))
        _derived = self._filtered_derived(derived_keys, "not-null" if ignore_null else "all")
        conditions = " AND ".join(f"{n} IS ?" for n, t, v in chain(self._mandatory, _derived))
        latest_col, latest_dir, latest_agg = self.latest_order_by()
        sql = (
            f"SELECT {fields} FROM {self.table} WHERE {conditions} ORDER BY {latest_col} {latest_dir} LIMIT 1"  # nosec
        )
        return sql

    def _get_latest(
        self,
        mandatory: Sequence[Any],
        derived: Optional[Dict[str, Any]] = None,
        ignore_null: bool = True,
        only: HashableContainer[str] = frozenset(),
        no: HashableContainer[str] = frozenset(),
    ) -> tuple:
        """Retrieve latest row based on mandatory and derived information.

        If `ignore_null=True` (default), values which are not provided in `derived`
        are ignored for the search.
        If it's `False`, values not provided in `derived` are assumed to be NULL
        and the row is only matches if the values in the database are NULL as well.

        The `only` argument can be used to include only these fields fields in the output.
        The `no` argument can be used to omit fields from the output.
        Only one of them can be specified.
        """

        derived = derived or {}
        sql = self._get_latest_sql(Keys(derived), ignore_null, only, no)
        args = self._args(mandatory, derived, ignore_null)
        self.cursor.execute(sql, args)
        return fetchone(self.cursor)

    def get_latest_many(
        self,
        mandatory: Iterable[Sequence[Any]],
        total: int,
        derived_names: Optional[Tuple[str, ...]] = None,
        derived: Optional[Iterable[Dict[str, Any]]] = None,
        ignore_null: bool = True,
        only: HashableContainer[str] = frozenset(),
        no: HashableContainer[str] = frozenset(),
    ) -> Iterator[tuple]:
        """Retrieve multiple latest rows based on mandatory and derived information.
        `derived_names` must be equal to the keys of the `derived` dicts.

        See `_get_latest` for more details.
        """

        # This function creates a join table to match multiple values at once
        # and also allow carrying the input order over to the output.
        # Another implementation could use `WHERE (x1 AND y1) OR (x2 AND y2)`,
        # but this way doesn't easily support sorting.

        derived_names = derived_names or ()
        derived = derived or {}

        _derived = self._filtered_derived(derived_names, "not-null" if ignore_null else "all")
        affected_fields = [n for n, t, v in chain(self._mandatory, _derived)]
        group_by = ", ".join(affected_fields)
        vars = len(affected_fields)
        values = ", ".join(f"({i}, {', '.join(repeat('?', vars))})" for i in range(total))
        select = ", ".join(f"t.{n}" for n in self._get_fields(only, no))
        join_on = " AND ".join(f"c.{n} = t.{n}" for n in affected_fields)
        join_group_by = ", ".join(f"t.{n}" for n in affected_fields)

        latest_col, latest_dir, latest_agg = self.latest_order_by()
        sql = f"""
            WITH conditions ({self.order_col}, {group_by}) AS (VALUES {values})  -- create table with order key and match values
            SELECT {select}
            FROM {self.table} t INNER JOIN conditions c ON {join_on}
            GROUP BY {join_group_by} HAVING {latest_col} = {latest_agg}({latest_col})  -- select only the latest entry from each group
            ORDER BY c.{self.order_col}  -- order rows according to input order
        """  # nosec

        args = tuple(chain.from_iterable(self._args_many(mandatory, derived, ignore_null)))

        self.cursor.execute(sql, args)

        return iterfetch(self.cursor)

    def _filtered_values_str(self, derived, ignore_null: bool):
        return

    def _add_file_query(self, derived_names: Optional[Tuple[str, ...]] = None, replace: bool = True) -> str:
        _derived = list(self._filtered_derived(derived_names, "not-null"))
        affected_fields = [n for n, t, v in chain(self._auto, self._mandatory, _derived)]
        fields = ", ".join(affected_fields)
        values = ", ".join(v for n, t, v in chain(self._auto, self._mandatory, _derived))

        if replace:
            sql = f"REPLACE INTO {self.table} ({fields}) VALUES ({values})"
        else:
            condition = " AND ".join(f"{n} IS excluded.{n}" for n, t, v in self._mandatory)
            set_affected = (f"{n}=excluded.{n}" for n in affected_fields)
            set_unaffected = (
                f"{n}=CASE WHEN {condition} THEN {n} ELSE NULL END"
                for n, t, v in self._filtered_derived(derived_names, "null")
            )
            update_set = ", ".join(chain(set_affected, set_unaffected))
            sql = (
                f"INSERT INTO {self.table} ({fields}) VALUES ({values}) ON CONFLICT DO UPDATE SET {update_set}"  # nosec
            )

        return sql

    def _add_file(self, mandatory: Sequence[Any], derived: Dict[str, Any], replace: bool = True) -> None:
        """Adds a new entry to the database and doesn't check if file
        with the same mandatory fields already exists.
        It will replace or update entries based on matching PRIMARY KEY or UNIQUE field, or else insert them.
        If `replace` is True the values previously existing in the existing replaced row but not given in `derived`
        will be overwritten with empty values.
        If `replace` is False, the given derived values will be updated if all mandatory fields match,
        and not given ones will keep their original values. If not all of the mandatory fields match,
        not given derived values will be reset to zero.  This makes sure out-dated values are removed.
        """

        sql = self._add_file_query(derived, replace)
        args = self._args(mandatory, derived, ignore_null=True)

        assert self.cursor.execute(sql, args).rowcount == 1

    def _add_file_many(
        self,
        mandatory: Iterable[Sequence[Any]],
        derived_names: Optional[Tuple[str, ...]] = None,
        derived: Optional[Iterable[Dict[str, Any]]] = None,
        replace: bool = True,
    ) -> None:
        """Replace or upsert multiple files.
        `derived_names` must be equal to the keys of the `derived` dicts.

        See `_add_file` for more details.
        """

        sql = self._add_file_query(derived_names, replace)
        args = self._args_many(mandatory, derived, ignore_null=True)
        self.cursor.executemany(sql, args)

    def _add_file_no_dup(
        self,
        mandatory: Sequence[Any],
        derived: Optional[Dict[str, Any]] = None,
        ignore_null: bool = True,
    ) -> bool:
        """Only adds a new entry to the db if the provided information
        does not exist in the db yet.
        """

        derived = derived or {}
        _derived = list(self._filtered_derived(derived, "not-null" if ignore_null else "all"))
        fields = ", ".join(n for n, t, v in chain(self._auto, self._mandatory, _derived))
        values = ", ".join(v for n, t, v in chain(self._auto, self._mandatory, _derived))
        conditions = " AND ".join(f"{n} IS ?" for n, t, v in chain(self._mandatory, _derived))

        sql = f"REPLACE INTO {self.table} ({fields}) SELECT {values} WHERE NOT EXISTS (SELECT 1 FROM {self.table} WHERE {conditions})"  # nosec
        args = self._args(mandatory, derived, ignore_null) * 2

        return self.cursor.execute(sql, args).rowcount == 1

        """ fixme: benchmark if this is really slower
        try:
            self._get_latest(mandatory, derived, ignore_null=ignore_null)
            return False

        except NoResult:
            self._add_file(mandatory, derived)
            return True
        """


class GenericFileDb(GenericDb):
    @classmethod
    def latest_order_by(cls) -> Tuple[str, str, str]:
        return ("entry_date", "DESC", "max")

    def get_latest(
        self,
        path: str,
        filesize: int,
        mod_date: int,
        derived: Optional[Dict[str, Any]] = None,
        ignore_null: bool = True,
        only: HashableContainer[str] = frozenset(),
        no: HashableContainer[str] = frozenset(),
    ) -> tuple:
        mandatory = (path, filesize, mod_date)
        return self._get_latest(mandatory, derived, ignore_null, only, no)

    def get(
        self, path: EntryType, only: HashableContainer[str] = frozenset(), no: HashableContainer[str] = frozenset()
    ) -> tuple:
        """Retrieves latest row based on mandatory information
        which is solely based on the `path`.
        Use `only`/`no` to include/exclude returned fields.
        """

        stats = path.stat()
        return self.get_latest(fspath(path), stats.st_size, stats.st_mtime_ns, ignore_null=True, only=only, no=no)

    def add(
        self, path: EntryType, derived: Optional[Dict[str, Any]] = None, commit: bool = True, replace: bool = True
    ) -> None:
        """Adds a new entry to the database and doesn't check if file
        with the same mandatory fields already exists.
        However it will replace entries based on PRIMARY KEYs or UNIQUE indices
        """

        derived = derived or {}
        stats = path.stat()
        mandatory = (fspath(path), stats.st_size, stats.st_mtime_ns)
        self._add_file(mandatory, derived, replace)
        if commit:
            self.commit()

    def add_file(self, path: str, filesize: int, mod_date: int, derived: Optional[Dict[str, Any]] = None) -> bool:
        """Only adds a new entry to the db if all the provided information doesn't match a row in the db yet."""

        mandatory = (path, filesize, mod_date)
        result = self._add_file_no_dup(mandatory, derived)
        self.commit()
        return result

    def add_files(self, batch: Iterable[Tuple[str, int, int, Optional[Dict[str, Any]]]]) -> Iterator[bool]:
        """Adds multiple new entries to the db while ignoring already existing entries with the same information.
        Yields True for new entries, False otherwise.
        """

        for path, filesize, mod_date, derived in batch:
            mandatory = (path, filesize, mod_date)
            yield self._add_file_no_dup(mandatory, derived)

        self.commit()

    def setdefault(self, path: EntryType, key: str, func: Callable[[], T]) -> T:
        try:
            (value,) = self.get(path, only=frozenset((key,)))
            logger.debug("Found %s of %s in db: %s", key, path, value)
        except NoResult:
            value = func()
            self.add(path, derived={key: value})
            logger.debug("Added %s of %s to db: %s", key, path, value)

        return value


class FileDbHistory(GenericFileDb):
    @classmethod
    def primary(cls):
        return [
            ("file_id", "INTEGER NOT NULL PRIMARY KEY", "?"),
        ]

    @classmethod
    def auto(cls):
        return [
            ("entry_date", "DATETIME", "datetime('now')"),
        ]

    @classmethod
    def mandatory(cls):
        return [
            ("path", "VARCHAR(256)", "?"),
            ("filesize", "INTEGER", "?"),
            ("mod_date", "INTEGER", "?"),
        ]

    def normalize_mandatory(self, mandatory: Sequence) -> Sequence:
        path, filesize, mod_date = mandatory
        drive, path = os.path.splitdrive(mandatory[0])
        path = normalize_seps(os.path.normpath(path))
        return path, filesize, mod_date


class FileDbSimple(GenericFileDb):
    @classmethod
    def primary(cls):
        return []

    @classmethod
    def auto(cls):
        return [
            ("entry_date", "DATETIME", "datetime('now')"),
        ]

    @classmethod
    def mandatory(cls):
        return [
            ("path", "VARCHAR(256) NOT NULL PRIMARY KEY", "?"),
            ("filesize", "INTEGER", "?"),
            ("mod_date", "INTEGER", "?"),
        ]


def is_signed_int_64(num: int) -> bool:
    return -(2**63) <= num <= 2**63 - 1


def unsigned_to_signed_int_64(num: int) -> int:
    return num - 2**63


def signed_to_unsigned_int_64(num: int) -> int:
    return num + 2**63


class Uint64:
    __slots__ = ("value",)

    def __init__(self, value: int) -> None:
        self.value = value

    def to_bytes(self):
        return self.value.to_bytes(length=8, byteorder="big", signed=False)


def uint64_to_bytes(uint64: Uint64) -> bytes:
    return uint64.to_bytes()


def uint64_from_bytes(blob: bytes) -> int:
    return int.from_bytes(blob, byteorder="big", signed=False)


class FileDbWithId(GenericFileDb):
    def __init__(self, dbpath: str, table: str, debug: bool = True, allow_add: bool = True) -> None:
        GenericFileDb.__init__(self, dbpath, table, debug, allow_add)
        sqlite3.register_adapter(Uint64, uint64_to_bytes)
        sqlite3.register_converter("uint64", uint64_from_bytes)

    @classmethod
    def primary(cls):
        return []

    @classmethod
    def auto(cls):
        return [
            ("entry_date", "DATETIME", "datetime('now')"),
        ]

    @classmethod
    def mandatory(cls):
        return [
            ("path", "VARCHAR(256) NOT NULL PRIMARY KEY", "?"),
            ("file_id", "uint64", "?"),
            ("device_id", "uint64", "?"),
            ("filesize", "INTEGER", "?"),
            ("mod_date", "INTEGER", "?"),
        ]

    def get(
        self, path: Path, only: HashableContainer[str] = frozenset(), no: HashableContainer[str] = frozenset()
    ) -> tuple:
        """Retrieves latest row based on mandatory information
        which is solely based on the `path`.
        Use `only`/`no` to include/exclude returned fields.
        """

        stats = path.stat()
        assert stats.st_ino != 0 and stats.st_dev != 0
        device = Uint64(stats.st_dev).to_bytes()
        inode = Uint64(stats.st_ino).to_bytes()
        mandatory = (fspath(path), inode, device, stats.st_size, stats.st_mtime_ns)
        return self._get_latest(mandatory, ignore_null=True, only=only, no=no)

    def add(
        self, path: Path, derived: Optional[Dict[str, Any]] = None, commit: bool = True, replace: bool = True
    ) -> None:
        """Adds a new entry to the database and doesn't check if file
        with the same mandatory fields already exists.
        However it will replace entries based on PRIMARY KEYs or UNIQUE indices
        """

        derived = derived or {}
        stats = path.stat()
        assert stats.st_ino != 0 and stats.st_dev != 0, stats

        # On windows st_dev and st_ino is unsigned 64 bit int, but sqlite only supports signed 64 bit ints.
        # I don't know about linux
        device = Uint64(stats.st_dev)
        inode = Uint64(stats.st_ino)

        mandatory = (fspath(path), inode, device, stats.st_size, stats.st_mtime_ns)
        self._add_file(mandatory, derived, replace)
        if commit:
            self.commit()



from typing import TYPE_CHECKING

from nltk.tokenize import word_tokenize

if TYPE_CHECKING:
    from os import PathLike


def count_words_in_file(path: PathLike, encoding: str = "utf-8") -> int:
    with open(path, encoding=encoding) as fr:
        return sum(len(word_tokenize(line)) for line in fr)



import csv
from typing import Dict, Iterator


def parse_imdb_csv(path: str) -> Iterator[Dict[str, str]]:
    with open(path, encoding="iso-8859-1", errors="strict", newline="") as fr:
        csvreader = csv.DictReader(fr)
        yield from csvreader



import logging
import sys
from string import Formatter
from traceback import format_exception
from typing import Any, Callable, Dict, Iterable, List, Optional, Sequence, Union

from rich.console import Console, Group, JustifyMethod, OverflowMethod, Style
from rich.markdown import Markdown
from rich.progress import BarColumn
from rich.progress import Progress as _RichProgress
from rich.progress import ProgressColumn, ProgressType, RenderableType
from rich.progress import Task as RichTask
from rich.progress import TaskID, TimeElapsedColumn
from rich.table import Column
from rich.text import Text
from typing_extensions import Self

from ._files import PathType
from .callbacks import BaseTask
from .callbacks import Progress as _Progress
from .file import copen

logger = logging.getLogger(__name__)


class NoneFormatter(Formatter):
    def __init__(self, default: Any) -> None:
        self.default = default
        super().__init__()

    def format_field(self, value, spec):
        if value is None:
            value = self.default
        return super().format_field(value, spec)


class DoubleFormatTextColumn(ProgressColumn):
    def __init__(
        self, text_format: str = "{task.description}", default: Any = 0.0, table_column: Optional[Column] = None
    ) -> None:
        self.text_format = text_format
        self.formatter = NoneFormatter(default)
        super().__init__(table_column=table_column or Column(no_wrap=True))

    def render(self, task: RichTask) -> Text:
        out = self.text_format.format(task=task)
        return Text(self.formatter.format(out, task=task))


def get_double_format_columns() -> List[ProgressColumn]:
    return [DoubleFormatTextColumn(), BarColumn(), TimeElapsedColumn()]


def install_markdown_excepthook(console: Optional[Console] = None) -> Callable:
    console_ = console or Console(file=sys.stderr)

    def excepthook(exc_type, exc_value, exc_traceback):
        lines = list(format_exception(exc_type, exc_value, exc_traceback))
        for line in lines[:-1]:
            console_.print(line, end="")

        args = [Markdown(arg) for arg in exc_value.args]
        if args:
            console_.print(f"{exc_type.__qualname__}:", *args)
        else:
            console_.print(f"{exc_type.__qualname__}")

    old_excepthook = sys.excepthook
    sys.excepthook = excepthook
    return old_excepthook


class Task(BaseTask):
    def __init__(
        self, progress: _RichProgress, total: Optional[float] = None, description: Optional[str] = None, **fields: Any
    ) -> None:
        self.progress = progress
        description = description or "Working..."
        self.task_id = self.progress.add_task(description, total=total, **fields)

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.progress.remove_task(self.task_id)

    def advance(self, delta: float) -> None:
        self.progress.advance(self.task_id, advance=delta)

    def update(
        self,
        *,
        completed: Optional[float] = None,
        total: Optional[float] = None,
        description: Optional[str] = None,
        **fields: Any,
    ) -> None:
        self.progress.update(self.task_id, completed=completed, total=total, description=description, **fields)


class Progress(_Progress):
    def __init__(self, progress: _RichProgress) -> None:
        self.progress = progress
        self.prolog: Optional[RenderableType] = None
        self.epilog: Optional[RenderableType] = None

        # patch `live.get_renderable` which is usually set to `rich.progress.Progress.get_renderable`
        self.progress.live.get_renderable = self.get_renderable

    def track(
        self,
        sequence: Union[Iterable[ProgressType], Sequence[ProgressType]],
        total: Optional[float] = None,
        description: Optional[str] = None,
        **fields: Any,
    ) -> Iterable[ProgressType]:
        description = description or "Working..."
        task_id = self.progress.add_task(description, total=total, **fields)
        try:
            yield from self.progress.track(sequence, task_id=task_id)
        finally:
            self.progress.remove_task(task_id)

    def track_auto(
        self,
        sequence: Union[Iterable[ProgressType], Sequence[ProgressType]],
        description: Optional[str] = None,
        **fields: Any,
    ) -> Iterable[ProgressType]:
        description = description or "Working..."
        with self.task(description=description, **fields) as task:
            for completed, item in enumerate(sequence):
                total = len(sequence)
                task.update(completed=completed, total=total)
                yield item

    def task(self, total: Optional[float] = None, description: str = "Working...", **fields: Any):
        return Task(self.progress, total, description, **fields)

    def get_renderable(self) -> RenderableType:
        renderables = []
        if self.prolog:
            renderables.append(self.prolog)
        if self.progress.tasks:
            renderables.append(self.progress.get_renderable())
        if self.epilog:
            renderables.append(self.epilog)

        return Group(*renderables)

    def set_prolog(self, prolog: Optional[RenderableType] = None) -> None:
        """Use to add arbitrary renderables before the progress to the live display.
        Pass None to disable.
        """

        self.prolog = prolog

    def set_epilog(self, epilog: Optional[RenderableType] = None) -> None:
        """Use to add arbitrary renderables after the progress to the live display.
        Pass None to disable.
        """

        self.epilog = epilog

    def print(self, s: str, end="\n") -> None:
        """Print a renderable to the console. It will not be part of the live/progress display,
        but printed to the normal console display. It will be done safely,
        i.e. not interrupting the progress display.
        """

        self.progress.print(
            s,
            end=end,
            markup=False,
            highlight=False,
            soft_wrap=True,
        )

    def refresh(self) -> None:
        self.progress.refresh()


class RichAriaProgress:
    """To be used with `genutility.aria.AriaDownloader` `active_callback` arguments."""

    def __init__(self, progress: _RichProgress) -> None:
        self.progress = progress
        self.tasks: Dict[str, TaskID] = {}  # gid to task_id

    def __call__(self, entries: Dict[str, Any]) -> None:
        existing = self.tasks.keys() & entries.keys()  # in both
        new = entries.keys() - self.tasks.keys()  # only in entries
        old = self.tasks.keys() - entries.keys()  # only in tasks

        for gid in existing:
            self.progress.update(
                self.tasks[gid],
                total=int(entries[gid]["totalLength"]),
                completed=int(entries[gid]["completedLength"]),
            )

        for gid in new:
            self.tasks[gid] = self.progress.add_task(
                gid,
                start=True,
                total=int(entries[gid]["totalLength"]),
                completed=int(entries[gid]["completedLength"]),
            )

        for gid in old:
            self.progress.remove_task(self.tasks[gid])
            del self.tasks[gid]


class StdoutFile:
    def __init__(
        self,
        console: Optional[Console] = None,
        path: Optional[PathType] = None,
        mode: str = "xt",
        encoding: Optional[str] = None,
        errors: Optional[str] = None,
        newline: Optional[str] = None,
        compresslevel: int = 9,
        style: Optional[Union[str, Style]] = None,
        justify: Optional[JustifyMethod] = None,
        overflow: Optional[OverflowMethod] = None,
        no_wrap: Optional[bool] = None,
        emoji: Optional[bool] = None,
        markup: Optional[bool] = None,
        highlight: Optional[bool] = None,
        width: Optional[int] = None,
        height: Optional[int] = None,
        crop: bool = True,
        soft_wrap: Optional[bool] = None,
        new_line_start: bool = False,
    ) -> None:
        if console is None and path is None:
            raise ValueError("console and path cannot both be None")

        assert mode in ("wt", "at", "xt")
        encoding = encoding or "utf-8"

        if path:
            self.fp = copen(path, mode, encoding=encoding, errors=errors, newline=newline, compresslevel=compresslevel)

            def _write(text: str) -> None:
                self.fp.write(text)

        else:
            self.fp = None

            def _write(text: str) -> None:
                console.print(
                    text,
                    end="",
                    style=style,
                    justify=justify,
                    overflow=overflow,
                    no_wrap=no_wrap,
                    emoji=emoji,
                    markup=markup,
                    highlight=highlight,
                    width=width,
                    height=height,
                    crop=crop,
                    soft_wrap=soft_wrap,
                    new_line_start=new_line_start,
                )

        self.write = _write

    def __enter__(self) -> Self:
        return self

    def close(self) -> None:
        if self.fp is not None:
            self.fp.close()

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


class StdoutFileNoStyle(StdoutFile):
    def __init__(self, *args, **kwargs) -> None:
        _kwargs = dict(
            encoding="utf-8",
            markup=False,
            highlight=False,
            soft_wrap=True,
        )
        _kwargs.update(kwargs)
        super().__init__(*args, **_kwargs)


class MarkdownHighlighter:
    def __call__(self, text: Union[str, Text]) -> Union[Text, Markdown]:
        if isinstance(text, str):
            return Markdown(text)
        elif isinstance(text, Text):
            return text
        else:
            raise TypeError(f"str or Text instance required, not {text!r}")


if __name__ == "__main__":
    install_markdown_excepthook()

    MARKDOWN = """
# This is an h1

Rich can do a pretty *decent* job of rendering markdown.

1. This is a list item
2. This is another list item
"""

    raise ValueError(MARKDOWN)



import re

import mistune

_markdown_newline_pat = re.compile(r"[^\n]\n[^\n]")


def _markdown_newline_func(x: re.Match) -> str:
    return x.group(0).replace("\n", "  \n")


def markdown_newline(s: str) -> str:
    """Convert newlines to markdown linebreaks."""

    return _markdown_newline_pat.sub(_markdown_newline_func, s)


_markdown_urls_pat = re.compile(
    r"(?<![\<])((?:http|https|ftp):\/\/[a-zA-Z0-9\-\.]+\.[a-zA-Z0-9]{2,}(?:\/[\:\/\?\#\[\]\@\!\$\&\'\(\)\*\+\;\=a-zA-Z0-9\-\._\~\%]*)?)(?![\>])"
)


def markdown_urls(s: str, ignore_trailing_dot: bool = True) -> str:
    """Converts text with URLs to text with markdown formatted URLs.
    markdown_urls("Visit https://google.com") -> "Visit <https://google.com>"
    This function is idempotent, which means that it will not convert URLs which are already in the correct format.
    markdown_urls("Visit <https://google.com>") -> "Visit <https://google.com>"
    """

    def markdown_dotfix(m):
        (url,) = m.groups()
        if ignore_trailing_dot and url.endswith("."):
            return f"<{url[:-1]}>."
        else:
            return f"<{url}>"

    # return _markdown_urls_pat.sub(r"<\1>", s)
    return _markdown_urls_pat.sub(markdown_dotfix, s)


class PlaintextRenderer(mistune.Renderer):
    """Mistune renderer to strip all markdown formatting and only return plaintext."""

    def block_code(self, code, lang=None):
        return code

    def block_quote(self, text):
        return text

    def block_html(self, html):
        return html

    def header(self, text, level, raw=None):
        return text + "\n\n"

    def hrule(self):
        return "\n"

    def list(self, body, ordered=True):
        return body + "\n"

    def list_item(self, text):
        return text + " "

    def paragraph(self, text):
        return text + "\n\n"

    def table(self, header, body):
        return body

    def table_row(self, content):
        return content

    def table_cell(self, content, **flags):
        return content

    # superscript is not supported!?

    def autolink(self, link, is_email=False):
        if is_email:
            return "<EMAIL>"
        else:
            return "<URL>"

    def codespan(self, text):
        return text

    def double_emphasis(self, text):
        return text

    def emphasis(self, text):
        return text

    def image(self, src, title, text):
        # text: alt attribute

        if title:
            return title
        return text

    def linebreak(self):
        return "\n"

    def newline(self):
        return "\n"

    def link(self, link, title, text):
        # link: href attribute
        if title:
            return title
        return text

    def strikethrough(self, text):
        return text

    def text(self, text):
        return text

    def inline_html(self, html):
        return html

    def escape(self, text):
        return text


def unescape(s: str) -> str:
    return s.replace("&lt;", "<").replace("&gt;", ">").replace("&amp;", "&")


md_text = mistune.Markdown(renderer=PlaintextRenderer())


def markdown2plaintext(s: str) -> str:
    """Converts markdown to plaintext."""

    return md_text.render(unescape(s)).strip()


md_html = mistune.Markdown(renderer=mistune.Renderer())


def markdown2html(s: str) -> str:
    """Converts markdown to HTML and strips surrounding <p> tags."""

    html = md_html.render(s).strip()
    if html.count("<p>") == 1 and html.count("</p>") == 1:
        return html[3:-4]  # assume p tags wrap everything
    else:
        return html



from os import PathLike, entrysuffix, fspath, scandir, wrap_text
from typing import IO, Callable, Dict, Iterator, Optional, Tuple, Union

from ._files import PathType
from .file import _check_arguments, _stripmode, copen


def iter_zip(
    file: Union[PathType, IO[bytes]],
    mode: str = "rb",
    encoding: Optional[str] = None,
    errors: Optional[str] = None,
    newline: Optional[str] = None,
    password: Optional[bytes] = None,
) -> Iterator[Tuple[str, IO]]:
    """
    Iterate file-pointers to archived files. They are valid for one iteration step each.
    If `file` is a file-like, it must be seekable. It is untested if unbuffered file-likes work.
    """

    # from pyzipper import AESZipFile as ZipFile
    from zipfile import ZipFile

    encoding = _check_arguments(mode, encoding)
    newmode = _stripmode(mode)

    with ZipFile(file, newmode) as zf:
        for zi in zf.infolist():
            if not zi.is_dir():
                with zf.open(zi, newmode, password) as bf:
                    yield zi.filename, wrap_text(bf, mode, encoding, errors, newline)


def iter_tar(
    file: Union[PathType, IO[bytes]],
    mode: str = "rb",
    encoding: Optional[str] = None,
    errors: Optional[str] = None,
    newline: Optional[str] = None,
) -> Iterator[Tuple[str, IO]]:
    """
    Iterate file-pointers to archived files. They are valid for one iteration step each.
    With the correct mode, `file` can be a non-seekable file-like. It is untested if unbuffered file-likes work.
    """

    import tarfile

    encoding = _check_arguments(mode, encoding)
    newmode = _stripmode(mode)

    if isinstance(file, PathLike):
        file = fspath(file)

    if isinstance(file, str):
        cls = tarfile.open(name=file, mode=newmode + "|*")
    else:
        cls = tarfile.open(fileobj=file, mode=newmode + "|*")

    with cls as tf:
        for ti in tf:
            if ti.isfile():  # fixme: what about links?
                bf = tf.extractfile(ti)
                assert bf, "member `ti` is file, but still `extractfile` returned `None`"
                with bf:  # no `mode` for `extractfile`
                    yield ti.name, wrap_text(bf, mode, encoding, errors, newline)


def iter_7zip(
    file: Union[PathType, IO[bytes]],
    mode: str = "rb",
    encoding: Optional[str] = None,
    errors: Optional[str] = None,
    newline: Optional[str] = None,
    password: Optional[str] = None,
) -> Iterator[Tuple[str, IO]]:
    """It is untested if unbuffered file-likes work."""

    from py7zr import SevenZipFile

    encoding = _check_arguments(mode, encoding)
    newmode = _stripmode(mode)

    with SevenZipFile(file, mode=newmode) as zf:
        for fname, bf in zf.readall().items():
            yield fname, wrap_text(bf, mode, encoding, errors, newline)


def iter_dir(
    path: PathType,
    mode: str = "rb",
    encoding: Optional[str] = None,
    errors: Optional[str] = None,
    newline: Optional[str] = None,
    follow_symlinks: bool = True,
    archives: bool = False,
    joiner: str = "/",
) -> Iterator[Tuple[str, IO]]:
    """
    Iterate file-pointers to files in directory `path`.
    They are valid for one iteration step each.
    If `archives` is True, archives like zip files will be traversed as well.
    """

    encoding = _check_arguments(mode, encoding)

    archive_funcs: Dict[
        str, Callable[[str, str, Optional[str], Optional[str], Optional[str]], Iterator[Tuple[str, IO]]]
    ] = {
        ".zip": iter_zip,
        ".tar": iter_tar,
        ".tgz": iter_tar,  # what about tar.gz
        ".tbz": iter_tar,  # what about tar.bz2
        ".7z": iter_7zip,
    }

    if archives:
        open = copen

    with scandir(path) as scan:
        for entry in scan:
            if entry.is_file(follow_symlinks=follow_symlinks):
                ext = entrysuffix(entry).lower()
                iter_archive = archive_funcs.get(ext, None)
                if archives and iter_archive:
                    for name, fr in iter_archive(entry.path, mode, encoding, errors, newline):
                        yield entry.path + joiner + name, fr
                else:
                    with open(entry.path, mode, encoding=encoding, errors=errors, newline=newline) as fr:
                        yield entry.path, fr

            elif entry.is_dir(follow_symlinks=follow_symlinks):
                yield from iter_dir(entry.path, mode, encoding, errors, newline, follow_symlinks, archives)



import sys
from collections.abc import Mapping, Sequence
from time import sleep
from typing import Any, Callable, Dict, Iterator, List
from typing import Mapping as MappingT
from typing import Optional
from typing import Sequence as SequenceT
from typing import Set, TextIO, Tuple, Union

from aria2p import Client
from aria2p.client import DEFAULT_HOST, DEFAULT_PORT, DEFAULT_TIMEOUT, ClientException
from requests.exceptions import ConnectionError

from .dict import update
from .exceptions import DownloadFailed, ExternalProcedureUnavailable, InconsistentState, WouldBlockForever, assert_type

CallbackFuncT = Callable[[Dict[str, Any]], None]


def aria_bool(value: Optional[bool]) -> Optional[str]:
    """aria2 rpc requires strings for some boolean arguments like 'continue'"""

    if value is None:
        return None
    elif value is True:
        return "true"
    elif value is False:
        return "false"
    else:
        raise ValueError(str(value))


def get_default_active_callback(file: TextIO = sys.stdout) -> CallbackFuncT:
    def default_active_callback(entries: Dict[str, Any]) -> None:
        completed = sum(int(entry["completedLength"]) for entry in entries.values())
        total = sum(int(entry["totalLength"]) for entry in entries.values())
        speed = sum(int(entry["downloadSpeed"]) for entry in entries.values())
        print(
            f"{len(entries)} downloads: {completed}/{total} bytes {speed} bytes/sec",
            file=file,
            end="\r",
        )

    return default_active_callback


def get_default_waiting_callback(file: TextIO = sys.stdout) -> CallbackFuncT:
    def default_waiting_callback(entries: Dict[str, Any]) -> None:
        print(f"{len(entries)} downloads waiting or paused", file=file, end="\r")

    return default_waiting_callback


class AriaDownloader:
    default_global_options = {
        "max-concurrent-downloads": 5,
        "remote-time": True,
    }

    default_options = {
        "max-connection-per-server": 1,
        "split": 5,
        "always-resume": True,
    }

    max_num_results = 100

    """ Download manager which uses aria2 instance to actually download the files.
        Tries to respect other users of the same instance and doesn't interfere with them.
    """

    def __init__(  # nosec
        self,
        host: str = DEFAULT_HOST,
        port: int = DEFAULT_PORT,
        secret: str = "",
        poll: float = 1.0,
        timeout: float = DEFAULT_TIMEOUT,
        global_options: Optional[dict] = None,
    ) -> None:
        """Initialize the aria2 client with `host`, `port` and `secret`.
        Poll aria2 every `poll` seconds.
        """

        self.aria2 = Client(host, port, secret, timeout)
        self.poll = poll

        if global_options:
            global_options = self.default_global_options.copy()
            global_options.update(global_options)
            self.aria2.change_global_option(global_options)
        else:
            self.aria2.change_global_option(self.default_global_options)

        self.gids: Set[str] = set()

    def query(self, method: str, *args: Any, **kwargs: Any) -> Any:
        try:
            return getattr(self.aria2, method)(*args, **kwargs)
        except ClientException as e:
            if e.code == 1:
                # either our code is bad, or some external actor removed our gid from aria
                raise InconsistentState(e.message)
            else:
                raise
        except ConnectionError as e:
            raise ExternalProcedureUnavailable(e)

    def pause_all(self) -> None:
        for gid in self.gids:
            self.query("pause", gid)

    def resume_all(self) -> None:
        for gid in self.gids:
            self.query("unpause", gid)

    def _entries(self, entries: Dict[str, Any]) -> Dict[str, Any]:
        return {entry["gid"]: entry for entry in entries if entry["gid"] in self.gids}

    def managed_downloads(self) -> int:
        return len(self.gids)

    def block_one(
        self, active_callback: Optional[CallbackFuncT] = None, waiting_callback: Optional[CallbackFuncT] = None
    ) -> Tuple[str, str]:
        """Blocks until one download is done.
        active_callback: Default `aria.get_default_active_callback()`
        waiting_callback: Default `aria.get_default_waiting_callback()`
        Can raise `requests.exceptions.ReadTimeout`
        """

        active_callback = active_callback or get_default_active_callback()
        waiting_callback = waiting_callback or get_default_waiting_callback()

        while True:
            """fixme: This loop has a (not very serious) race condition.

            For example a download might change its status from waiting to active
            during the two queries. Then it would not be found.
            This cannot be fixed by changing the query order as the status change graph
            (waiting<->active->stopped) cannot be sorted topologically.
            And even if you could, returning stopped downloads has priority over
            waiting/active ones, as otherwise we would block for too long.

            Does using a multi/batch-call fix this?
            """

            entries = self._entries(self.query("tell_stopped", 0, self.max_num_results))  # complete or error
            if entries:
                gid, entry = entries.popitem()

                try:
                    if entry["status"] == "complete":
                        assert len(entry["files"]) == 1
                        return gid, entry["files"][0]["path"]
                    elif entry["status"] == "error":
                        raise DownloadFailed(gid, entry["errorCode"], entry["errorMessage"])
                    elif entry["status"] == "removed":
                        raise RuntimeError("Someone removed our download...")
                    else:
                        raise RuntimeError("Unexpected status: {}".format(entry["status"]))

                finally:
                    self.remove_stopped(gid)

            entries = self._entries(self.query("tell_active"))  # active
            if entries:
                active_callback(entries)
                sleep(self.poll)
                continue

            entries = self._entries(self.query("tell_waiting", 0, self.max_num_results))  # waiting or paused
            if entries:
                waiting_callback(entries)
                sleep(self.poll)
                continue

            if self.gids:
                """Actually only this check is sensitive to race condition, as the looping logic
                would care of retrying otherwise.
                However this is the only way to check for external modifications.
                """
                raise InconsistentState(
                    "Some downloads got lost. We either encountered a race condition \
                    or some external actor removed the download"
                )

            raise WouldBlockForever("No downloads active or waiting")

    def block_all_iter(
        self, active_callback: Optional[CallbackFuncT] = None, waiting_callback: Optional[CallbackFuncT] = None
    ) -> Iterator[Tuple[Optional[Exception], Optional[str], Optional[str]]]:
        while True:
            try:
                gid, path = self.block_one(active_callback, waiting_callback)
                yield (None, gid, path)
            except WouldBlockForever:
                return
            except DownloadFailed as e:
                yield (e, None, None)

    def block_all(
        self, active_callback: Optional[CallbackFuncT] = None, waiting_callback: Optional[CallbackFuncT] = None
    ) -> List[Tuple[Optional[Exception], Optional[str], Optional[str]]]:
        return list(self.block_all_iter(active_callback, waiting_callback))

    def remove_stopped(self, gid: str) -> None:
        self.gids.remove(gid)

        # removes a complete/error/removed download
        # fails on active/waiting/paused (on CTRL-C for example)
        self.query("remove_download_result", gid)

    def remove(self, gid: str) -> None:
        self.gids.remove(gid)
        self.query("remove", gid)

    def block_gid(self, gid: str, progress_file: Optional[TextIO] = sys.stdout) -> str:
        """Blocks until download is done.
        If progress printing is not needed, `progress_file` should be set to `None`.
        Returns the path of the downloaded file on disk.
        """

        if gid not in self.gids:
            raise KeyError("Invalid GID")

        try:
            while True:
                s = self.query("tell_status", gid)

                status = s["status"]

                if status == "active":
                    if progress_file:
                        print(
                            s["completedLength"],
                            "/",
                            s["totalLength"],
                            "bytes",
                            s["downloadSpeed"],
                            "bytes/sec",
                            file=progress_file,
                            end="\r",
                        )

                elif status == "waiting":
                    if progress_file:
                        print("waiting", file=progress_file, end="\r")

                elif status == "paused":
                    if progress_file:
                        print("paused", file=progress_file, end="\r")

                elif status == "error":
                    raise DownloadFailed(
                        gid, s["errorCode"], s["errorMessage"]
                    )  # RuntimeError: No URI available. errorCode=8 handle

                elif status == "complete":
                    assert len(s["files"]) == 1
                    return s["files"][0]["path"]

                elif status == "removed":
                    raise RuntimeError("Someone removed our download...")

                else:
                    raise RuntimeError(f"Unexpected status: {status}")

                sleep(self.poll)
        finally:
            self.remove_stopped(gid)

    def download(
        self,
        uri: str,
        path: Optional[str] = None,
        filename: Optional[str] = None,
        headers: Optional[Union[SequenceT[str], MappingT[str, str]]] = None,
        max_connections: Optional[int] = None,
        split: Optional[int] = None,
        continue_: Optional[bool] = None,
        retry_wait: Optional[int] = None,
        max_tries: Optional[int] = None,
        connect_timeout: Optional[int] = None,
        timeout: Optional[int] = None,
        no_netrc: Optional[bool] = None,
    ) -> str:
        """Downloads `uri` to directory `path`.
        Does not block. Returns a download identifier.
        """

        if headers is not None:
            assert_type("headers", headers, (Sequence, Mapping))

        if isinstance(headers, Mapping):
            headers = [f"{k}: {v}" for k, v in headers.items()]

        options = self.default_options.copy()
        update(
            options,
            {
                "dir": path,
                "out": filename,
                "header": headers,
                "max-connection-per-server": max_connections,
                "split": split,
                "continue": aria_bool(continue_),
                "retry-wait": retry_wait,
                "max-tries": max_tries,
                "connect-timeout": connect_timeout,
                "timeout": timeout,
                "no-netrc": aria_bool(no_netrc),
            },
        )

        gid = self.query("add_uri", [uri], options)
        self.gids.add(gid)
        return gid

    def block_x(
        self,
        num: int,
        active_callback: Optional[CallbackFuncT] = None,
        waiting_callback: Optional[CallbackFuncT] = None,
    ) -> Tuple[Optional[str], Optional[str]]:
        if self.managed_downloads() >= num:
            finished_gid, path = self.block_one(active_callback, waiting_callback)
            return finished_gid, path
        else:
            return None, None

    def download_x(
        self,
        num: int,
        uri: str,
        path: Optional[str] = None,
        filename: Optional[str] = None,
        headers: Optional[Union[SequenceT[str], MappingT[str, str]]] = None,
        max_connections: Optional[int] = None,
        split: Optional[int] = None,
        continue_: Optional[bool] = None,
        retry_wait: Optional[int] = None,
        max_tries: Optional[int] = None,
        connect_timeout: Optional[int] = None,
        timeout: Optional[int] = None,
        no_netrc: Optional[bool] = None,
        active_callback: Optional[CallbackFuncT] = None,
        waiting_callback: Optional[CallbackFuncT] = None,
    ) -> Tuple[str, Optional[str], Optional[str]]:
        queued_gid = self.download(
            uri,
            path,
            filename,
            headers,
            max_connections,
            split,
            continue_,
            retry_wait,
            max_tries,
            connect_timeout,
            timeout,
            no_netrc,
        )
        finished_gid, path = self.block_x(num, active_callback, waiting_callback)
        return queued_gid, finished_gid, path


class DownloadManager:
    """To use from a single thread.

    State transitions:
    waiting -> running (automatically)
    running -> paused (manually)
    paused -> running (manually)
    running -> complete (automatically)
    running -> error (automatically)
    """

    def __init__(self):
        pass

    def download_with_callback(self, url, path=None, filename=None, headers=None, force=False, func=None):
        """Starts or enqueues download and a unique id for that download job.
        if `force` is True, the download is started without regard for the current number
        of downloads and queue position.
        """

    def block_uid(self, uid):
        """Blocks until `uid` completes or errors
        and returns information about that download.
        """

    def block_one(self):
        """Block until any download completes or errors
        and returns information about that download.
        """

    def block_active(self, x):
        """Block if more than `x` downloads are currently active
        until x or less are active. Yields information about completed/errored downloads
        as they become available.
        """

    def cancel(self, uid, states=None):
        """Cancels a download identified by `uid`. if `states` is None, it will be
        - removed from the queue if waiting
        - canceled if downloading
        - removed if completed
        - removed if errored.
        The previous state will be returned,
        """

        # sets the status if `uid` to canceled.

    def clean(self, uid, states=None):
        """Removed all files associated to `uid` from disk. `states` is a list of acceptable states.
        For example ("stopped", "error") will only remove files if the download
        was stopped mid-progress or because of an error.
        Take notice that ("complete") will remove files of completed downloads from disk.
        """

        # doesn't change status

    def forget(self, uid):
        """Remove `uid` from the manager completely. It can only do so for downloads
        with the state "canceled".
        """

    # for convenience only

    def block_all(self, uid):
        return self.block_active(0)

    def download_and_block(self, url):
        uid = self.download_with_callback(url)
        return self.block_uid(uid)


if __name__ == "__main__":
    from argparse import ArgumentParser

    parser = ArgumentParser()
    parser.add_argument("action", choices=("download", "remove"))
    parser.add_argument("--uris", metavar="URI", nargs="+", default=[], help="URLs to download")
    parser.add_argument("--outpath", default=".", help="Output directory")
    parser.add_argument("--max", default=2, type=int, help="Maximum concurrent downloads")
    parser.add_argument("--gids", metavar="GID", nargs="+", default=[], help="GIDs")
    args = parser.parse_args()

    d = AriaDownloader()
    if args.action == "download":
        for uri in args.uris:
            try:
                queued_gid, finished_gid, path = d.download_x(args.max, uri, args.path)
                print(f"Downloaded {uri} to {path}")
            except DownloadFailed as e:
                print("Download failed: %s", e)

        for a in d.block_all():
            print(a)

    elif args.action == "remove":
        for gid in args.gids:
            d.query("remove", gid)



from typing import Tuple

DAR_WS = 16 / 9
DAR_FS = 4 / 3

RES_PAL = (720, 576)
RES_NTSC = (720, 480)
RES_PAL_BLANKING = (704, 576)
RES_NTSC_BLANKING = (704, 480)
RES_PAL_HALF_H = (352, 576)
RES_NTSC_HALF_H = (352, 480)
RES_HALF_HD = (1280, 720)
RES_FULL_HD = (1920, 1080)

PAR_HD = 1


def SAR(RES: Tuple[int, int]) -> float:
    x, y = RES
    return x / y


# Generic / Digital PAR
PAR_PAL_16_9 = DAR_WS / SAR(RES_PAL)  # 64:45
PAR_PAL_4_3 = DAR_FS / SAR(RES_PAL)  # 16:15
PAR_PAL_4_3_BLANKING = DAR_FS / SAR(RES_PAL_BLANKING)  # 12:11
PAR_PAL_HALF_H = DAR_FS / SAR(RES_PAL_HALF_H)  # 24:11
PAR_NTSC_16_9 = DAR_WS / SAR(RES_NTSC)  # 32:27
PAR_NTSC_4_3 = DAR_FS / SAR(RES_NTSC)  # 8:9
PAR_NTSC_4_3_BLANKING = DAR_FS / SAR(RES_NTSC_BLANKING)  # 10:11
PAR_NTSC_HALF_H = DAR_FS / SAR(RES_NTSC_HALF_H)  # 20:11

# MPEG-4 PAR
PAR_MPG4_PAL_16_9 = PAR_PAL_OVERSCAN_16_9 = DAR_WS / SAR(RES_PAL_BLANKING)  # 16:11
PAR_MPG4_PAL_4_3 = PAR_PAL_OVERSCAN_4_3 = DAR_FS / SAR(RES_PAL_BLANKING)  # 12:11
PAR_MPG4_NTSC_16_9 = PAR_NTSC_OVERSCAN_16_9 = DAR_WS / SAR(RES_NTSC_BLANKING)  # 40:33
PAR_MPG4_NTSC_4_3 = PAR_NTSC_OVERSCAN_4_3 = DAR_FS / SAR(RES_NTSC_BLANKING)  # 10:11

# ITU-R BT.601 PAR (Rec. 601)
PAR_ITU_PAL_16_9 = 512 / 351
PAR_ITU_PAL_4_3 = 128 / 117
PAR_ITU_NTSC_16_9 = 5760 / 4739
PAR_ITU_NTSC_4_3 = 4320 / 4739

# Frames per seconds
REAL_FILM_FPS = 24
NTSC_PROG_FPS = 24000 / 1001
NTSC_VIDEO_FPS = 30000 / 1001
PAL_PROG_FPS = 25
PAL_VIDEO_FPS = 30



import numpy as np
from scipy.linalg import solve_triangular
from scipy.optimize import linear_sum_assignment

# from .numba import opjit


def linear_sum_assignment_cost(cost_matrix):
    """Minimum weight matching in bipartite graphs using the Hungarian/Kuhn-Munkres algorithm."""

    row_ind, col_ind = linear_sum_assignment(cost_matrix)
    return np.sum(cost_matrix[row_ind, col_ind])


# @opjit(["f4[:,:](f4[:,:,:], f4[:,:], b1, b1, b1)", "f8[:,:](f8[:,:,:], f8[:,:], b1, b1, b1)"])
def batch_solve_triangular(
    matrices: np.ndarray,
    vectors: np.ndarray,
    lower: bool = False,
    overwrite_b: bool = False,
    check_finite: bool = False,
) -> np.ndarray:
    B, M1, M2 = matrices.shape
    B2, V = vectors.shape

    if B != B2 or M1 != M2 or M2 != V:
        raise ValueError("Input dimensions are invalid")

    out = np.empty_like(vectors)

    for i in range(B):
        out[i] = solve_triangular(
            matrices[i], vectors[i], lower=lower, overwrite_b=overwrite_b, check_finite=check_finite
        )

    return out


if __name__ == "__main__":
    import timeit

    from .numpy import random_triangular_matrix

    size = 100
    bs = 200
    Ms = np.array([random_triangular_matrix(size, lower=True) for _ in range(bs)])
    vs = np.random.uniform(0, 1, (bs, size))

    batch_solve_triangular(Ms, vs, lower=True)  # warmup
    print(min(timeit.repeat("batch_solve_triangular(Ms, vs, lower=True)", number=10, repeat=5, globals=globals())))



from operator import itemgetter
from typing import Callable, Dict, Iterator, MutableSequence, Optional, Sequence, Tuple, TypeVar

from .func import identity
from .indexing import triangular_indices
from .typing import Comparable, Orderable

T = TypeVar("T")


class LazyStringList:
    def __init__(self, length: int) -> None:
        self.length = length

    def __getitem__(self, idx: int) -> str:
        return str(idx)

    def __len__(self) -> int:
        return self.length

    def __iter__(self) -> Iterator[str]:
        return map(str, range(self.length))

    def __contains__(self, idx: int) -> bool:
        return 0 <= idx < self.length


def pop_many(seq: MutableSequence[T], func: Callable) -> Iterator[T]:
    """`pop()`s values from `seq` where func(value) is true.
    For performance reasons the elements are popped
    and yielded in reverse order.
    """

    idx = [i for i, elm in enumerate(seq) if func(elm)]

    for i in reversed(idx):
        yield seq.pop(i)


def triangular(seq: Sequence[T]) -> Iterator[Tuple[T, T]]:
    """Returns all combinations of items of `seq` with duplicates and self-combinations.
    triangular([1, 2, 3, 4]) -> (1, 2), (1, 3), (1, 4), (2, 3), (2, 4), (3, 4)
    """

    for i, j in triangular_indices(len(seq)):
        yield seq[i], seq[j]


def sliding_window(seq: Sequence[T], size: int, step: int = 1) -> Iterator[Sequence[T]]:
    """Similar to `iter.x_wise()` except that it returns slices of a sequence.
    sliding_window([1, 2, 3], 2) -> [1, 2], [2, 3]
    sliding_window([1, 2, 3, 4, 5], 2, 3) -> [1, 2], [4, 5]
    """

    seqlen = len(seq)
    if size > seqlen:
        raise ValueError("size must be less or equal the length of the sequence")

    for i in range(0, seqlen - size + 1, step):
        yield seq[i : i + size]


def batch(seq: Sequence[T], size: int) -> Iterator[Sequence[T]]:
    """Similar to `iter.batch()` except that it returns slices of a sequence.
    batch([1, 2, 3], 2) -> [1, 2], [3]
    batch([1, 2, 3, 4, 5], 3) -> [1, 2, 3], [4, 5]
    """

    if size < 1:
        raise ValueError("size must be larger than 0")

    seqlen = len(seq)

    for i in range(0, (seqlen + size - 1) // size):
        yield seq[i * size : (i + 1) * size]


def delete_duplicates_from_sorted_sequence(
    seq: MutableSequence[T], key: Optional[Callable[[T], Comparable]] = None
) -> None:
    """Deletes duplicates from a sorted sequence `seq` based on `key`.
    Works in-place.
    """

    key = key or identity
    i = 0
    while i < len(seq) - 1:
        if key(seq[i]) == key(seq[i + 1]):
            del seq[i + 1]
        else:
            i += 1


def remove_duplicate_rows_from_sorted_table(seq: MutableSequence[Dict[str, T]], key: str) -> None:
    return delete_duplicates_from_sorted_sequence(seq, itemgetter(key))


def cycle_sort(seq: MutableSequence[Orderable]) -> Iterator[Tuple[int, int]]:
    """Sort a sequence in place and yield swaps.
    based on: https://rosettacode.org/wiki/Sorting_algorithms/Cycle_sort#Python
    """

    # Loop through the sequence to find cycles to rotate.
    for cycle_start, item in enumerate(seq):
        # Find where to put the item.
        pos = cycle_start
        for item2 in seq[cycle_start + 1 :]:
            if item2 < item:
                pos += 1

        # If the item is already there, this is not a cycle.
        if pos == cycle_start:
            continue

        # Otherwise, put the item there or right after any duplicates.
        while item == seq[pos]:
            pos += 1
        seq[pos], item = item, seq[pos]
        yield cycle_start, pos

        # Rotate the rest of the cycle.
        while pos != cycle_start:
            # Find where to put the item.
            pos = cycle_start
            for item2 in seq[cycle_start + 1 :]:
                if item2 < item:
                    pos += 1

            # Put the item there or right after any duplicates.
            while item == seq[pos]:
                pos += 1
            seq[pos], item = item, seq[pos]
            if pos != cycle_start:
                yield cycle_start, pos



from functools import partial
from typing import Any, Iterable, List

from pyspark.sql import functions as f
from pyspark.sql import types as t
from pyspark.sql.column import Column
from pyspark.sql.dataframe import DataFrame


def timestamp_from_unixtime_ms(colname: str) -> Column:
    """Probably requires Spark>=3.0.0"""

    dt = f.from_unixtime(f.col(colname) / 1000).cast("timestamp")
    msec = f.expr("INTERVAL 1 milliseconds") * (f.col(colname) % 1000)
    return (dt + msec).alias(colname)


def null_columns(df: DataFrame) -> List[str]:
    """Return a list of column names of columns which are only ever have null."""

    df_count = df.count()
    null_counts = df.select([f.count(f.when(f.col(c).isNull(), c)).alias(c) for c in df.columns]).collect()[0].asDict()
    return [k for k, v in null_counts.items() if v >= df_count]


def drop_null_columns(df: DataFrame) -> DataFrame:
    """Drop all columns from dataframe which are only ever null."""

    to_drop = null_columns(df)
    return df.drop(*to_drop)


def lit_array(it: Iterable[Any]) -> Column:
    """Create a Pyspark array of literal values."""

    return f.array(list(map(f.lit, it)))


sparkmap = {
    "int64": t.LongType,
    "uint64": partial(t.DecimalType, 20, 0),
    "int32": t.IntegerType,
    "uint32": t.LongType,
    "str": t.StringType,
    "bool": t.BooleanType,
}


def schema_simple_to_spark(d: dict, sort_keys: bool = False) -> t.StructType:
    fields = []

    if sort_keys:
        keys: Iterable = sorted(d)
    else:
        keys = d

    for k in keys:
        if isinstance(d[k], dict):
            fields.append(t.StructField(k, schema_simple_to_spark(d[k], sort_keys)))
        elif isinstance(d[k], list):
            if d[k]:
                if isinstance(d[k][0], dict):
                    fields.append(t.StructField(k, t.ArrayType(schema_simple_to_spark(d[k][0], sort_keys))))
                else:
                    fields.append(t.StructField(k, t.ArrayType(sparkmap[d[k][0]]())))
        else:
            fields.append(t.StructField(k, sparkmap[d[k]]()))

    return t.StructType(fields)



import torch
from torch.distributions.categorical import Categorical

_ZERO_SIZE = torch.Size()


def categorical(probs: torch.tensor, size: torch.tensor = _ZERO_SIZE) -> torch.tensor:
    return Categorical(probs).sample(size)



from typing import Callable, List, Optional


class SegmentTree:
    # ported from https://codeforces.com/blog/entry/18051

    t: List[Optional[int]]

    def __init__(self, arr: List[int], func: Callable[[int, int], int], initializer: int) -> None:
        self.n = len(arr)
        self.func = func
        self.initializer = initializer

        self.t = [None] * self.n + arr

    def build(self) -> None:
        i = self.n - 1
        while i > 0:
            self.t[i] = self.func(self.t[i << 1], self.t[i << 1 | 1])
            i -= 1

    def modify(self, p: int, value: int) -> None:
        p += self.n
        self.t[p] = value

        while p > 1:
            val = self.func(self.t[p], self.t[p ^ 1])
            p >>= 1
            self.t[p] = val

    def query(self, left: int, right: int) -> int:
        if left >= right or not 0 <= left <= self.n or not 0 <= right <= self.n:
            raise ValueError(f"Interval [{left}, {right}) out of range")

        res = self.initializer
        left += self.n
        right += self.n

        while left < right:
            if left & 1:
                res = self.func(res, self.t[left])
                left += 1
            if right & 1:
                right -= 1
                res = self.func(res, self.t[right])

            left >>= 1
            right >>= 1

        return res



import csv
from functools import partial
from itertools import islice
from typing import Callable, Iterator, List, Optional, Sequence

from .dict import itemgetter
from .func import compose, identity, zipmap


def iter_csv(path: str, delimiter: str = ",", encoding: str = "utf-8", skip: int = 0) -> Iterator[List[str]]:
    with open(path, encoding="utf-8", newline="") as fr:
        yield from islice(csv.reader(fr, delimiter=delimiter), skip, None)


def read_csv(
    path: str,
    delimiter: str = ",",
    skip: int = 0,
    usecols: Optional[Sequence[int]] = None,
    dtype: Optional[Sequence[Callable]] = None,
    encoding: str = "utf-8",
) -> Iterator[tuple]:
    with open(path, encoding=encoding, newline="") as csvfile:
        if usecols:
            if dtype:
                getcols = compose(partial(zipmap, dtype), itemgetter(usecols))
            else:
                getcols = itemgetter(usecols)
        else:
            if dtype:
                getcols = partial(zipmap, dtype)
            else:
                getcols = identity

        for row in islice(csv.reader(csvfile, delimiter=delimiter), skip, None):
            print(row)
            yield tuple(getcols(row))



from lxml import etree  # nosec


def xml_xslt_to_xhtml(path_xml: str, path_xslt: str, path_xhtml: str) -> None:
    """Only use with trusted xml data"""

    xml = etree.parse(path_xml)  # nosec
    xslt = etree.parse(path_xslt)  # nosec
    transform = etree.XSLT(xslt)
    newdom = transform(xml)
    with open(path_xhtml, "wb") as xhtml:
        newdom.write(xhtml, pretty_print=True, xml_declaration=True, encoding="utf-8")



from operator import itemgetter

import spacy

from .metrics import same_words_similarity
from .text_segmentation import split_paragraphs


def get_spacy_tokenizer(modelname, lemmatize=True):
    nlu = spacy.load(modelname)

    if lemmatize:
        return lambda text: (t.lemma_ for t in nlu(text))
    else:
        return lambda text: map(str, nlu(text))


class NoParagraphsFound(Exception):
    pass


class QueryBasedParagraphExtraction:
    def __init__(self, modelname="en_core_web_sm"):
        self.tokenizer = get_spacy_tokenizer(modelname)

    def extract(self, text: str, query: str) -> str:
        paragraphs = split_paragraphs(text)
        it = ((same_words_similarity(p, query, self.tokenizer), p) for p in paragraphs)

        try:
            return max(it, key=itemgetter(0))[1]
        except ValueError:
            raise NoParagraphsFound()


from .test import MyTestCase


class TextSummarizationTests(MyTestCase):
    text = [
        """Copyright laws are changing all over the world. Be sure to check the
copyright laws for your country before downloading or redistributing
this or any other Project Gutenberg eBook.
""",
        """This header should be the first thing seen when viewing this Project
Gutenberg file. Please do not remove it. Do not change or edit the
header without written permission.
""",
        """Please read the "legal small print," and other information about the
eBook and Project Gutenberg at the bottom of this file. Included is
important information about your specific rights and restrictions in
how the file may be used. You can also find out about how to make a
donation to Project Gutenberg, and how to get involved.
""",
    ]

    def test_a(self):
        fulltext = "\n".join(self.text)

        pe = QueryBasedParagraphExtraction("en_core_web_sm")

        result = pe.extract(fulltext, "LAWS")
        self.assertEqual(self.text[0], result)

        result = pe.extract(fulltext, "legal law involve")
        self.assertEqual(self.text[2], result)


if __name__ == "__main__":
    import unittest

    unittest.main()



from inspect import currentframe, getfile, getouterframes, getsourcefile
from pathlib import Path
from sys import _getframe  # sys._getframe() is not guaranteed to exist
from typing import Sequence, Union


def get_calling_functions(level: Union[int, Sequence[int]] = 2, sep: str = "->") -> str:
    curframe = currentframe()
    if curframe is None:
        raise RuntimeError("This Python interpreter does not provide stack frame support")

    calframe = getouterframes(curframe)
    del curframe

    if isinstance(level, int):
        return f"{Path(calframe[level].filename).stem}:{calframe[level].function}"
    else:
        return sep.join(f"{Path(calframe[level].filename).stem}:{calframe[level].function}" for level in level)


def caller_file(depth: int = 1) -> Path:
    return Path(getfile(_getframe(depth))).resolve().parent


def caller_sourcefile(depth: int = 1) -> Path:
    return Path(getsourcefile(_getframe(depth))).resolve().parent



from typing import Any, Dict, Tuple, Union

import piexif
from PIL import Image, ImageDraw, ImageFont
from PIL.ExifTags import GPSTAGS, TAGS

from .exceptions import NoActionNeeded

Color = Union[str, Tuple[int, int, int]]


def multiline_textsize(text: str, ttf: ImageFont, spacing: int = 4) -> Tuple[int, int]:
    lines = text.splitlines()

    width = 0
    height = (len(lines) - 1) * spacing

    for line in lines:
        # _, _, w, h = ttf.getbbox(line)
        left, top, right, bottom = ttf.getbbox(line)
        # assert left == 0, left
        # assert top == 0, top
        width = max(width, right)
        height = height + bottom

    return width, height


def exifinfo(image: Image.Image) -> Dict[str, Any]:
    ret: Dict[str, Any] = {}

    exifd = image._getexif()

    if exifd is None:
        return {}

    for k1, v1 in exifd.items():
        try:
            tag = TAGS[k1]
            if tag == "GPSInfo":
                ret[tag] = {}
                for k2, v2 in v1.items():
                    try:
                        gpstag = GPSTAGS[k2]
                        ret[tag][gpstag] = v2
                    except KeyError:
                        pass
            else:
                ret[tag] = v1
        except KeyError:
            pass

    return ret


def text_with_outline(
    draw: ImageDraw,
    pos: Tuple[int, int],
    text: str,
    font: ImageFont,
    fillcolor: Color,
    outlinecolor: Color,
    outlinesize: int = 1,
) -> None:
    x, y = pos
    delta = outlinesize

    # thin border
    draw.text((x - delta, y), text, font=font, fill=outlinecolor)
    draw.text((x + delta, y), text, font=font, fill=outlinecolor)
    draw.text((x, y - delta), text, font=font, fill=outlinecolor)
    draw.text((x, y + delta), text, font=font, fill=outlinecolor)

    # thicker border
    draw.text((x - delta, y - delta), text, font=font, fill=outlinecolor)
    draw.text((x + delta, y - delta), text, font=font, fill=outlinecolor)
    draw.text((x - delta, y + delta), text, font=font, fill=outlinecolor)
    draw.text((x + delta, y + delta), text, font=font, fill=outlinecolor)

    # now draw the text over it
    draw.text((x, y), text, font=font, fill=fillcolor)


def write_text(
    img: Image.Image,
    text: str,
    alignment: str = "TL",
    fillcolor: Color = (255, 255, 255),
    outlinecolor: Color = (0, 0, 0),
    fontsize: Union[float, int] = 0.03,
    padding: Union[float, Tuple[int, int]] = (5, 5),
) -> None:
    if alignment not in {"TL", "TC", "TR", "BL", "BC", "BR"}:
        raise ValueError(f"Invalid alignment: {alignment}")

    if isinstance(fontsize, int):
        pass
    elif isinstance(fontsize, float):
        fontsize = int(img.height * fontsize)
    else:
        raise TypeError("fontsize must be float or int")

    if isinstance(padding, tuple):
        pass
    elif isinstance(padding, float):
        padding = int(img.width * padding), int(img.height * padding)
    else:
        raise TypeError("padding must be float or Tuple[int, int]")

    font = ImageFont.truetype("arial.ttf", fontsize)

    d = ImageDraw.Draw(img)
    size_text = d.textsize(text, font=font)

    if alignment == "TL":
        pos = padding
    elif alignment == "TC":
        pos = (img.width // 2 - size_text[0] // 2, padding[1])
    elif alignment == "TR":
        pos = (img.width - padding[0] - size_text[0], padding[1])
    elif alignment == "BL":
        pos = (padding[0], img.height - padding[1] - size_text[1])
    elif alignment == "BC":
        pos = (img.width // 2 - size_text[0] // 2, img.height - padding[1] - size_text[1])
    elif alignment == "BR":
        pos = (img.width - padding[0] - size_text[0], img.height - padding[1] - size_text[1])

    text_with_outline(d, pos, text, font, fillcolor, outlinecolor, 2)


def _fix_orientation(img: Image.Image, orientation: int) -> Image.Image:
    if orientation == 1:
        raise NoActionNeeded("File already properly rotated")
    elif orientation == 2:
        img = img.transpose(Image.Transpose.FLIP_LEFT_RIGHT)
    elif orientation == 3:
        img = img.transpose(Image.Transpose.ROTATE_180)
    elif orientation == 4:
        img = img.transpose(Image.Transpose.FLIP_TOP_BOTTOM)
    elif orientation == 5:
        img = img.transpose(Image.Transpose.TRANSPOSE)
    elif orientation == 6:
        img = img.transpose(Image.Transpose.ROTATE_270)
    elif orientation == 7:
        img = img.transpose(Image.Transpose.TRANSVERSE)
    elif orientation == 8:
        img = img.transpose(Image.Transpose.ROTATE_90)
    else:
        raise ValueError(f"Unsupported orientation: {orientation}")

    return img


def fix_orientation(img: Image.Image, exif: dict) -> Image.Image:
    orientation = exif["0th"][piexif.ImageIFD.Orientation]
    img = _fix_orientation(img, orientation)
    exif["0th"][piexif.ImageIFD.Orientation] = 1

    return img



from fractions import Fraction
from typing import Tuple, Union

import numpy as np

from .numpy import bincount_batch, center_of_mass_2d, unblock

Rational = Union[float, Fraction]


def grayscale(arr: np.ndarray) -> np.ndarray:
    """Converts RGB/BGR images `arr` [batch..., x, y, channel] to grayscale."""

    return np.sum(arr, axis=-1) // arr.shape[-1]


def histogram_1d(arr: np.ndarray, levels: int) -> np.ndarray:
    """Input shape of `arr`: [batch..., x]. Histogrammed over x and batched over the remaining dimensions.
    Output shape: [batch..., levels]
    """

    return bincount_batch(arr, -1, levels)


def resize_oar(max_width: int, max_height: int, dar: Rational) -> Tuple[int, int]:
    maxdar = max_width / max_height

    if dar >= maxdar:  # wider than it should be
        width = max_width
        height = int(width / dar)
    else:  # thinner than it should be
        height = max_height
        width = int(height * dar)

    return width, height


def resize_maxsize(max_width: int, max_height: int, width: int, height: int) -> Tuple[int, int]:
    return resize_oar(max_width, max_height, width / height)


def histogram_2d(arr: np.ndarray, levels: int) -> np.ndarray:
    """Input shape of `arr`: [batch..., y, x]. Histogrammed over x and y and batched over
    the remaining dimensions.
    Output shape: [batch..., levels]
    """

    if len(arr.shape) < 2:
        raise ValueError("arr must be at least 2-dimensional")

    newshape = arr.shape[:-2] + (arr.shape[-2] * arr.shape[-1],)
    flattened = np.reshape(arr, newshape)

    return bincount_batch(flattened, -1, levels)


def block_histogram_2d(arr: np.ndarray, by: int, bx: int, levels: int) -> np.ndarray:
    """Input shape of `arr`: [batch..., y, x]. Histogrammed over blocks of size bx and by
    and batched over the remaining dimensions.
    Output shape: [batch..., y/by, x/bx, levels]
    """

    invx = arr.shape[-1] // bx  # dimensions in unblock go from innerst to outerst
    invy = arr.shape[-2] // by
    blocks = unblock(arr, invx, invy)
    block_hists = histogram_1d(blocks, levels)
    return block_hists.reshape(arr.shape[:-2] + (invy, invx, -1))


def image_histogram(arr: np.ndarray, levels: int = 256) -> np.ndarray:
    """Input shape of RGB/BGR image `arr`: [batch..., x, y, channel]. It is summed over channels to create a grayscale
    image, then histogrammed over x and y and batched over the remaining dimensions.
    Output shape: [batch..., levels]
    """

    if len(arr.shape) < 3:
        raise ValueError("arr must be at least 3-dimensional")

    gray = grayscale(arr)
    return histogram_2d(gray, levels)


def image_block_histogram(arr: np.ndarray, bx: int, by: int, levels: int = 256) -> np.ndarray:
    """Input shape of RGB/BGR image `arr`: [batch..., x, y, channel]. It is summed over channels to create a grayscale
    image, then histogrammed over x and y and batched over the remaining dimensions.
    Output shape: [batch..., bx, by, levels]
    """

    if len(arr.shape) < 3:
        raise ValueError("arr must be at least 3-dimensional")

    gray = grayscale(arr)
    return block_histogram_2d(gray, bx, by, levels)


def center_of_mass_quadrant(img: np.ndarray) -> int:
    """Returns the image quadrant of the center of mass of the grayscale image `img`:
    +---+---+
    | 0 | 1 |
    +---+---+
    | 3 | 2 |
    +---+---+

    If the center of mass is exactly in the center or the total mass is 0, it will return quadrant 2,
    because of *how less* than and *greater than* conditions are used.
    """

    if len(img.shape) != 2:
        raise ValueError("Grayscale image expected")

    cm_y, cm_x = center_of_mass_2d(img)
    d_y, d_x = img.shape
    c_y = (d_y - 1) / 2
    c_x = (d_x - 1) / 2

    if cm_y < c_y:
        if cm_x < c_x:
            quadrant = 0
        else:
            quadrant = 1
    else:
        if cm_x < c_x:
            quadrant = 3
        else:
            quadrant = 2

    return quadrant


def normalize_image_rotation(img: np.ndarray, img_gray: np.ndarray) -> np.ndarray:
    quadrant = center_of_mass_quadrant(img_gray)
    return np.rot90(img, k=quadrant, axes=(0, 1))



"""Scripts using line_profiler.profile can import this to be able to run unmodified
without the profiler present.
"""

import logging

try:
    profile = profile  # bind so it can be imported
    logging.info("Running profiler")

except NameError:

    def profile(func):
        return func



import codecs
import logging
import os.path
import platform
import re
import sqlite3
from typing import Iterable, List, Optional
from urllib.parse import urlencode

from ._files import to_dos_path
from .callbacks import Progress
from .iter import batch
from .signal import safe_for_loop
from .string import build_multiple_replace

_is_win = platform.system() == "Windows"


def compile_options() -> List[str]:
    with sqlite3.connect(":memory:") as conn:
        cur = conn.execute("PRAGMA compile_options;")
        ret = cur.fetchall()
    return [i[0] for i in ret]


def quote_identifier(s: str, errors: str = "strict") -> str:
    """Escapes sqlite identifiers (like table or column names).
    Copied from: https://stackoverflow.com/a/6701665
    """

    encodable = s.encode("utf-8", errors).decode("utf-8")
    nul_index = encodable.find("\x00")

    if nul_index >= 0:
        error = UnicodeEncodeError("NUL-terminated utf-8", encodable, nul_index, nul_index + 1, "NUL not allowed")
        error_handler = codecs.lookup_error(errors)
        replacement, _ = error_handler(error)
        encodable = encodable.replace("\x00", replacement)

    return '"' + encodable.replace('"', '""') + '"'


def vacuum(db_path: str) -> None:
    """Vacuums a sqlite database."""

    with sqlite3.connect(db_path) as conn:
        conn.execute("VACUUM")


def batch_executer(
    cursor: sqlite3.Cursor,
    query_str: str,
    it: Iterable[tuple],
    batch_size: int = 10000,
    exclusive: bool = True,
    progress: Optional[Progress] = None,
) -> int:
    """Execute `query_str` with parameters from `it` batch-wise with batches of size `batch_size`.
    If `exclusive` is True the database will be locked in exclusive mode.
    """

    if exclusive:
        cursor.execute("PRAGMA locking_mode=EXCLUSIVE")  # might speed things up

    entries = 0

    progress = progress or Progress()
    for data in batch(progress.track(it), batch_size, list):
        # need to cache batch data, because if the iterable is exhausted,
        # executemany raises `sqlite3.ProgrammingError`
        if data:
            cursor.execute("BEGIN TRANSACTION")
            cursor.executemany(query_str, data)
            cursor.execute("COMMIT TRANSACTION")
            entries += len(data)

    return entries


def safe_batch_executer(
    cursor: sqlite3.Cursor,
    query_str: str,
    it: Iterable[tuple],
    batch_size: int = 10000,
    exclusive: bool = True,
    progress: Optional[Progress] = None,
) -> None:
    """Execute `query_str` with parameters from `it` batch-wise with batches of size `batch_size`.
    If `exclusive` is True the database will be locked in exclusive mode.
    If an db integrity error occurs, the batch will be skipped and the transaction for
    this batch rolled back.
    If the iterable is advanced, the user can be sure its elements will be inserted into the
    database, even if a KeyboardInterrupt is received in-between.
    """

    if exclusive:
        cursor.execute("PRAGMA locking_mode=EXCLUSIVE")  # might speed things up

    progress = progress or Progress()
    source = batch(progress.track(it), batch_size)

    def sqlexec(queries_batch):
        try:
            cursor.execute("BEGIN TRANSACTION")
            cursor.executemany(query_str, queries_batch)
            cursor.execute("COMMIT TRANSACTION")
        except sqlite3.IntegrityError:
            logging.info("Skipping batch")
            cursor.execute("ROLLBACK TRANSACTION")
        except sqlite3.OperationalError:
            raise

    try:
        safe_for_loop(source, sqlexec)
    except KeyboardInterrupt:
        logging.info("Batch execution safely interrupted")


_percent_encode = build_multiple_replace({"%": "%25", "?": "%3f", "#": "%23"})


def to_uri(path: str, **params: str) -> str:
    # https://www.sqlite.org/uri.html#the_uri_path

    query = urlencode(params)
    if _is_win:
        path = to_dos_path(path)
        is_abs = os.path.isabs(path)
    path = _percent_encode(path)
    if _is_win:
        path = re.sub("[\\/]+", "/", path)
    else:
        path = re.sub("/+", "/", path)
    if _is_win and is_abs:
        path = "/" + path
    return f"file:{path}?{query}"



from typing import Iterator

import numpy as np

from .videofile import AvVideo, BadFile  # noqa: F401


class CorruptFile(Exception):
    pass


def iter_video(path: str) -> Iterator[np.ndarray]:
    """Yield frames from a video file. For video device support see `genutility.cv.iter_video`."""

    with AvVideo(path) as video:
        for _time, frame in video.iterall(native=True):
            if frame.is_corrupt:
                raise CorruptFile("Frame %s at %s of %s is corrupt", frame.index, frame.time, path)
            yield frame.to_ndarray()



import logging

from typing_extensions import Self

from ..exceptions import MalformedFile

REAL_FILM_FPS = 24.0
NTSC_PROG_FPS = 24000.0 / 1001
NTSC_VIDEO_FPS = 30000.0 / 1001
PAL_PROG_FPS = 25.0
PAL_VIDEO_FPS = 30.0

DEFAULT_ENCODING = "utf-8-sig"


def ntsc_to_pal(num):
    return num * NTSC_PROG_FPS / PAL_PROG_FPS


def pal_to_ntsc(num):
    return num * PAL_PROG_FPS / NTSC_PROG_FPS


def film_to_ntsc(num):
    return num * REAL_FILM_FPS / NTSC_PROG_FPS


def to_msec(string):
    h, m, s, ms = (int(i) for i in string.replace(",", ":").replace(".", ":").split(":"))
    return (h * 60 * 60 + m * 60 + s) * 1000 + ms


def to_srt_time(t):
    r, ms = divmod(t, 1000)
    r, s = divmod(r, 60)
    h, m = divmod(r, 60)
    return f"{int(h):02d}:{int(m):02d}:{int(s):02d},{int(ms):03d}"


class Subtitle:
    def __init__(self):
        self.num = 0
        self.start = 0
        self.end = 0
        self.lines = []

    def get_times(self):
        return to_srt_time(self.start), to_srt_time(self.end)

    def set_times(self, start, end):
        self.start, self.end = to_msec(start), to_msec(end)

    def append(self, line):
        self.lines.append(line)


class SRTFile:
    nl = "\n"
    sep = " --> "

    def __init__(self, filename, mode="r", encoding=DEFAULT_ENCODING, overwrite_index=False):
        self.state = 0
        self.linenum = 0
        self.sub_num = 0
        self.fp = open(filename, mode, encoding=encoding, errors="replace")
        self.overwrite_index = overwrite_index

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def close(self):
        self.fp.close()

    def write_subtitle(self, subtitle):
        start, end = subtitle.get_times()
        if self.overwrite_index:
            self.fp.write(str(self.sub_num) + self.nl)
        else:
            self.fp.write(str(subtitle.num) + self.nl)
        self.fp.write(start + self.sep + end + self.nl)
        for line in subtitle.lines:
            self.fp.write(line)
            self.fp.write(self.nl)
        self.fp.write(self.nl)
        self.sub_num += 1

    def read_line(self):
        self.linenum += 1
        line = next(self.fp).rstrip()

        if self.state == 0:
            try:
                self.sub.num = int(line)
            except ValueError:
                raise MalformedFile(f"Error in line {self.linenum}: srt malformed: {line}")
            self.state = 1
        elif self.state == 1:
            start, end = line.split(self.sep)
            self.sub.set_times(start, end)
            self.state = 2
        elif self.state == 2:
            if line == "":
                self.state = 0
            else:
                self.sub.append(line)

    def read_subtitle(self):
        self.sub = Subtitle()
        self.read_line()
        while self.state != 0:
            self.read_line()  # can throw, and leave incomplete subtitle in buffer (eg. if file doesnt end in newline)
        return self.sub

    def __iter__(self):
        if self.fp:
            return self
        else:
            raise ValueError("I/O operation on closed file.")

    def __next__(self):
        return self.read_subtitle()


def transform(infile, outfile, callback, encoding=DEFAULT_ENCODING):
    with SRTFile(infile, "r", encoding) as fi, SRTFile(outfile, "w", encoding) as fo:
        for subtitle in fi:
            subtitle = callback(subtitle)
            fo.write_subtitle(subtitle)


def merge(srtlines, lim):
    ret = []
    for line in srtlines:
        if line.startswith(lim):
            ret.append(line[len(lim) :])
        else:
            if not ret:
                ret.append(line)
            else:
                ret[-1] = ret[-1] + " " + line
    return ret


def srt2txt(srt_fp):
    lim = "- "
    pos = 0
    lines = []

    while True:
        if pos >= len(lines):
            sub = next(srt_fp)
            lines = merge(sub.lines, lim)
            pos = 0
        ret = lines[pos]
        pos += 1
        yield ret


def compare_srt_and_txt(srt_file, txt_file):
    with SRTFile(srt_file, "r", encoding="utf-8-sig") as srt, open(txt_file, encoding="utf-8-sig") as txt:
        srtiter = srt2txt(srt)

        limit = 10

        while limit > 0:
            try:
                line = next(txt).rstrip()
                if not line:
                    continue
                if not line.startswith("- "):
                    print(line)
                else:
                    line = line[2:]
                    start = 0
                    while start <= len(line):
                        srtpart = next(srtiter)
                        txtpart = line[start : start + len(srtpart)]
                        logging.debug(f"Compare '{txtpart}' '{srtpart}'")
                        if txtpart != srtpart:
                            # print("{} - {}".format(sub.start, sub.end))
                            yield (txtpart, srtpart)
                            limit -= 1
                        start += len(srtpart) + 1
            except StopIteration:
                break



from twisted.internet import protocol


class GenericFactory(protocol.Factory):
    def __init__(self, protocol, *args, **kwargs):
        self.protocol = protocol
        self.args = args
        self.kwargs = kwargs

    def buildProtocol(self, addr):
        return self.protocol(self, *self.args, **self.kwargs)  # deleted "self, " here...
